from bizyengine.core import BizyAirBaseNode, data_types
from bizyengine.core.nodes_io import BizyAirNodeIO


class CLIPTextEncodeFlux(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "clip": (data_types.CLIP,),
                "clip_l": ("STRING", {"multiline": True, "dynamicPrompts": True}),
                "t5xxl": ("STRING", {"multiline": True, "dynamicPrompts": True}),
                "guidance": (
                    "FLOAT",
                    {"default": 3.5, "min": 0.0, "max": 100.0, "step": 0.1},
                ),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    RETURN_NAMES = ("conditioning",)
    # FUNCTION = "encode"

    CATEGORY = "advanced/conditioning/flux"

    # def encode(self, clip: BizyAirNodeIO, clip_l: str, t5xxl: str, guidance: float):
    #     new_clip = clip.copy(self.assigned_id)
    #     new_clip.add_node_data(
    #         class_type="CLIPTextEncodeFlux",
    #         inputs={
    #             "clip": clip,
    #             "clip_l": clip_l,
    #             "t5xxl": t5xxl,
    #             "guidance": guidance,
    #         },
    #         outputs={"slot_index": 0},
    #     )
    #     return (new_clip,)


class FluxGuidance(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conditioning": (data_types.CONDITIONING,),
                "guidance": (
                    "FLOAT",
                    {"default": 3.5, "min": 0.0, "max": 100.0, "step": 0.1},
                ),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    FUNCTION = "append"

    CATEGORY = "advanced/conditioning/flux"

    def append(self, conditioning: BizyAirNodeIO, guidance):
        new_conditioning = conditioning.copy(self.assigned_id)
        new_conditioning.add_node_data(
            class_type="FluxGuidance",
            inputs={
                "conditioning": conditioning,
                "guidance": guidance,
            },
            outputs={"slot_index": 0},
        )
        return (new_conditioning,)



import json
import pprint

import bizyengine.core as bizyair
from aiohttp import web
from bizyengine.core.data_types import BIZYAIR_TYPE_MAP
from server import PromptServer


def get_bizyair_display_name(class_type: str) -> str:
    bizyair_cls_prefix = bizyair.nodes_base.PREFIX
    bizyair_logo = bizyair.nodes_base.LOGO
    return f"{bizyair_logo}{bizyair_cls_prefix} {bizyair.NODE_DISPLAY_NAME_MAPPINGS.get(class_type, class_type)}"


def workflow_convert(inputs: dict):
    nodes = inputs["nodes"]
    for node in nodes:
        class_type = node["type"]
        node_inputs = node.get("inputs")
        node_outputs = node.get("outputs")
        bizyair_cls_type = f"{bizyair.nodes_base.PREFIX}_{class_type}"
        is_converted = False
        if bizyair_cls_type in bizyair.NODE_CLASS_MAPPINGS:
            node["type"] = bizyair_cls_type
            display_name = get_bizyair_display_name(class_type)
            node["properties"]["Node name for S&R"] = display_name
            if node_inputs:
                for input_node in node_inputs:
                    input_type = input_node["type"]
                    input_node["type"] = BIZYAIR_TYPE_MAP.get(input_type, input_type)
            if node_outputs:
                for output_node in node_outputs:
                    output_type = output_node["type"]
                    output_node["type"] = BIZYAIR_TYPE_MAP.get(output_type, output_type)
            is_converted = True
        pprint.pprint(
            {
                "original_class_type": class_type,
                "bizyair_cls_type": bizyair_cls_type,
                "is_converted": is_converted,
            }
        )
    return inputs


@PromptServer.instance.routes.post("/bizyair/node_converter")
async def convert(request):
    try:
        data = await request.json()
        ret = workflow_convert(data)
        return web.Response(
            text=json.dumps(ret),
            content_type="application/json",
        )
    except Exception as e:
        return web.json_response({"status": "error", "message": str(e)}, status=400)



import importlib
import logging
import warnings
from functools import wraps
from typing import List

from .data_types import is_send_request_datatype
from .nodes_io import BizyAirNodeIO, create_node_data

try:
    comfy_nodes = importlib.import_module("nodes")
except ModuleNotFoundError:
    warnings.warn("Importing comfyui.nodes failed!")
    comfy_nodes = type("nodes", (object,), {"NODE_DISPLAY_NAME_MAPPINGS": {}})

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

LOGO = "☁️"
PREFIX = f"BizyAir"
NODE_CLASS_MAPPINGS = {}
NODE_DISPLAY_NAME_MAPPINGS = {}


def to_camel_case(s):
    return "".join(word.capitalize() for word in s.split("_"))


def validate_category(cls, prefix):
    assert cls.CATEGORY.startswith(f"☁️{prefix}")


def register_node(cls, prefix):
    class_name = (
        f"{prefix}_{cls.__name__}"
        if not cls.__name__.startswith(prefix)
        else cls.__name__
    )
    logger.debug(
        f"Class: {cls}, Name: {class_name}, Has DISPLAY_NAME: {hasattr(cls, 'NODE_DISPLAY_NAME')}"
    )

    if hasattr(cls, "NODE_DISPLAY_NAME"):
        display_name = cls.NODE_DISPLAY_NAME
        if not display_name.startswith(f"{LOGO}{prefix}"):
            display_name = f"{LOGO}{prefix} {display_name}"
    else:
        base_name = class_name[len(prefix) + 1 :]
        if base_name in comfy_nodes.NODE_DISPLAY_NAME_MAPPINGS:
            display_name = (
                f"{LOGO}{prefix} {comfy_nodes.NODE_DISPLAY_NAME_MAPPINGS[base_name]}"
            )
        else:
            display_name = f"{LOGO}{prefix} {base_name}"
            logger.warning(
                f"Display name '{display_name}' might differ from the native display name."
            )

    NODE_CLASS_MAPPINGS[class_name] = cls
    NODE_DISPLAY_NAME_MAPPINGS[class_name] = display_name


def ensure_unique_id(org_func, original_has_unique_id=False):
    @wraps(org_func)
    def new_func(self, **kwargs):
        if original_has_unique_id:
            self._assigned_id = kwargs.get("unique_id", "UNIQUE_ID")
        elif "unique_id" in kwargs:
            self._assigned_id = kwargs.pop("unique_id")
        return org_func(self, **kwargs)

    return new_func


def ensure_hidden_unique_id(org_input_types_func):
    original_has_unique_id = False

    @wraps(org_input_types_func)
    def new_input_types_func():
        nonlocal original_has_unique_id

        result = org_input_types_func()
        if "hidden" not in result:
            result["hidden"] = {"unique_id": "UNIQUE_ID"}
        elif "unique_id" not in result["hidden"]:
            result["hidden"]["unique_id"] = "UNIQUE_ID"
        else:
            original_has_unique_id = True
        return result

    # Ensure original_has_unique_id is correctly set before returning
    new_input_types_func()
    return new_input_types_func, original_has_unique_id


class BizyAirBaseNode:
    FUNCTION = "default_function"

    def __init_subclass__(cls, **kwargs):
        if not cls.CATEGORY.startswith(f"{LOGO}{PREFIX}"):
            cls.CATEGORY = f"{LOGO}{PREFIX}/{cls.CATEGORY}"
        register_node(cls, PREFIX)
        cls.setup_input_types()

    @classmethod
    def setup_input_types(cls):
        # https://docs.comfy.org/essentials/custom_node_more_on_inputs#hidden-inputs
        new_input_types_func, original_has_unique_id = ensure_hidden_unique_id(
            cls.INPUT_TYPES
        )
        cls.INPUT_TYPES = new_input_types_func
        setattr(
            cls,
            cls.FUNCTION,
            ensure_unique_id(getattr(cls, cls.FUNCTION), original_has_unique_id),
        )

    @property
    def assigned_id(self):
        assert self._assigned_id is not None and isinstance(self._assigned_id, str)
        return str(self._assigned_id)

    def default_function(self, **kwargs):
        class_type = self._determine_class_type()

        node_ios = self._process_non_send_request_types(class_type, kwargs)
        # TODO: add processing for send_request_types
        send_request_datatype_list = self._get_send_request_datatypes()
        if len(send_request_datatype_list) == len(self.RETURN_TYPES):
            return self._process_all_send_request_types(node_ios)
        return node_ios

    def _get_send_request_datatypes(self):
        return [
            return_type
            for return_type in self.RETURN_TYPES
            if is_send_request_datatype(return_type)
        ]

    def _determine_class_type(self):
        class_type = getattr(self, "CLASS_TYPE_NAME", type(self).__name__)
        if class_type.startswith(f"{PREFIX}_"):
            class_type = class_type[len(PREFIX) + 1 :]
        return class_type

    def _process_non_send_request_types(self, class_type, kwargs):
        outs = []
        for slot_index, _ in enumerate(self.RETURN_TYPES):
            node = BizyAirNodeIO(node_id=self.assigned_id, nodes={})
            node.add_node_data(
                class_type=class_type, inputs=kwargs, outputs={"slot_index": slot_index}
            )
            outs.append(node)
        return tuple(outs)

    def _process_all_send_request_types(self, node_ios: List[BizyAirNodeIO]):
        out = node_ios[0].send_request()
        assert len(out) == len(self.RETURN_TYPES)
        return out



from aiohttp import web

from .errno import ErrorNo, errnos


def JsonResponse(http_status_code, data):
    return web.json_response(
        data,
        status=http_status_code,
        content_type="application/json",
    )


def OKResponse(data):
    return JsonResponse(
        200, {"message": "success", "code": errnos.OK.code, "data": data}
    )


def ErrResponse(err: ErrorNo):
    return JsonResponse(
        err.http_status_code,
        {"message": err.message, "code": err.code, "data": err.data},
    )



import glob
import json
import os
import time
from abc import ABC, abstractmethod
from collections import OrderedDict
from dataclasses import dataclass
from typing import Any, Dict


@dataclass
class CacheConfig:
    max_size: int = 100
    expiration: int = 300  # 300 seconds
    cache_dir: str = "./cache"
    file_prefix: str = "bizyair_cache_"
    file_suffix: str = ".json"
    use_cache: bool = True

    @classmethod
    def from_config(cls, config: Dict[str, Any]):
        return cls(
            max_size=config.get("max_size", 100),
            expiration=config.get("expiration", 300),
            cache_dir=config.get("cache_dir", "./cache"),
            file_prefix=config.get("file_prefix", "bizyair_cache_"),
            file_suffix=config.get("file_suffix", ".json"),
            use_cache=config.get("use_cache", True),
        )


class CacheManager(ABC):
    @abstractmethod
    def get(self, key):
        pass

    @abstractmethod
    def set(self, key, value):
        pass

    @abstractmethod
    def clear(self):
        pass

    @abstractmethod
    def disable(self):
        pass


class BizyAirTaskCache(CacheManager):
    def __init__(self, config: CacheConfig):
        self.config = config
        self.cache = OrderedDict()
        self.cache_dir = config.cache_dir
        self.ensure_directory_exists()
        self.cache = self.load_cache() if config.use_cache else self.cache

    def ensure_directory_exists(self):
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

    def load_cache(self):
        cache_v_files = glob.glob(
            os.path.join(
                self.cache_dir, f"{self.config.file_prefix}*{self.config.file_suffix}"
            )
        )
        output = OrderedDict()
        cache_datas = []
        for cache_file in cache_v_files:
            try:
                file_name = os.path.basename(cache_file)[
                    len(self.config.file_prefix) : -len(self.config.file_suffix)
                ]
                cache_key = file_name.split("-")[0]
                cache_timestamp = file_name.split("-")[1]
                if int(time.time()) - int(cache_timestamp) > self.config.expiration:
                    self.delete_file(cache_file)
                    continue
                cache_datas.append(
                    {
                        "key": cache_key,
                        "timestamp": int(cache_timestamp),
                        "file_path": cache_file,
                    }
                )
            except Exception as e:
                print(
                    f"Warning: Error loading cache file {cache_file}: because {e}, will delete it"
                )
        cache_datas = sorted(cache_datas, key=lambda x: x["timestamp"])
        for cache_data in cache_datas:
            output[cache_data["key"]] = (
                cache_data["file_path"],
                cache_data["timestamp"],
            )
        return output

    def delete(self, key):
        if key in self.cache:
            self.delete_file(self.cache[key][0])
            del self.cache[key]

    def get(self, key):
        if key not in self.cache:
            return None

        file_path, timestamp = self.cache[key]
        if time.time() - timestamp >= self.config.expiration:
            self._remove_expired_entry(file_path, key)
            return None

        cache_data = self._read_file(file_path)
        if cache_data["cache_key"] == key:
            return cache_data["result"]
        else:
            self._remove_expired_entry(file_path, key)
            return None

    def _read_file(self, file_path):
        try:
            with open(file_path, "r") as f:
                cache_data = json.load(f)
                return cache_data
        except Exception as e:
            print(f"Error reading file {file_path}: {e}")
        return None

    def _remove_expired_entry(self, file_path, key):
        self.delete_file(file_path)
        del self.cache[key]

    def set(self, key, value, *, overwrite=False):
        if not overwrite and key in self.cache:
            raise ValueError(
                f"Key '{key}' already exists in cache. Use overwrite=True to replace it."
            )
        assert isinstance(key, str), "Key must be a string"

        if len(self.cache) >= self.config.max_size:
            self._evict_oldest()

        timestamp = int(time.time())
        file_path = os.path.join(
            self.cache_dir, f"{self.config.file_prefix}{key}-{timestamp}.json"
        )
        self.write_file(key, value, file_path, timestamp)

    def _evict_oldest(self):
        oldest_key, (oldest_file_path, _) = self.cache.popitem(last=False)
        self.delete_file(oldest_file_path)

    def write_file(self, key: str, value: Any, file_path: str, timestamp: int):
        try:
            with open(file_path, "w") as f:
                json.dump(
                    {"result": value, "cache_key": key, "timestamp": timestamp}, f
                )
            self.cache[key] = (file_path, timestamp)
        except Exception as e:
            print(f"Error writing file for key '{key}': {e}")

    def delete_file(self, file_path):
        if os.path.exists(file_path):
            try:
                os.remove(file_path)
            except Exception as e:
                print(f"Error deleting file '{file_path}': {e}")

    def clear(self):
        for file_path, _ in self.cache.values():
            self.delete_file(file_path)
        self.cache.clear()

    def disable(self):
        self.clear()


# Example usage
if __name__ == "__main__":
    cache_config = CacheConfig(max_size=12, expiration=10, cache_dir="./cache")
    cache = BizyAirTaskCache(cache_config)

    # Set some cache values
    cache.set("key1", "This is the value for key1")
    cache.set("key2", "This is the value for key2")

    # Retrieve values from cache
    print(cache.get("key1"))  # Output: This is the value for key1
    print(cache.get("key2"))  # Output: This is the value for key2

    # Wait for expiration
    time.sleep(9)
    print(cache.get("key1"))  # Output: None (expired)

    # Clear cache
    cache.clear()
    print(cache.get("key2"))  # Output: None (cache cleared)



# ComfyUI/comfy_extras/nodes_model_advanced.py
import nodes
from bizyengine.core import BizyAirBaseNode, BizyAirNodeIO, data_types


class ModelSamplingSD3(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (data_types.MODEL,),
                "shift": (
                    "FLOAT",
                    {"default": 3.0, "min": 0.0, "max": 100.0, "step": 0.01},
                ),
            }
        }

    RETURN_TYPES = (data_types.MODEL,)
    # FUNCTION = "patch"

    CATEGORY = "advanced/model"


class ModelSamplingFlux(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (data_types.MODEL,),
                "max_shift": (
                    "FLOAT",
                    {"default": 1.15, "min": 0.0, "max": 100.0, "step": 0.01},
                ),
                "base_shift": (
                    "FLOAT",
                    {"default": 0.5, "min": 0.0, "max": 100.0, "step": 0.01},
                ),
                "width": (
                    "INT",
                    {
                        "default": 1024,
                        "min": 16,
                        "max": nodes.MAX_RESOLUTION,
                        "step": 8,
                    },
                ),
                "height": (
                    "INT",
                    {
                        "default": 1024,
                        "min": 16,
                        "max": nodes.MAX_RESOLUTION,
                        "step": 8,
                    },
                ),
            }
        }

    RETURN_TYPES = (data_types.MODEL,)
    # FUNCTION = "patch"
    CATEGORY = "advanced/model"



from bizyengine.core import BizyAirBaseNode, data_types
from bizyengine.core.path_utils import path_manager as folder_paths


class InstantIDModelLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "instantid_file": (folder_paths.get_filename_list("instantid"),)
            }
        }

    RETURN_TYPES = (data_types.INSTANTID,)
    FUNCTION = "default_function"
    CATEGORY = "InstantID"
    NODE_DISPLAY_NAME = "Load InstantID Model"


class ApplyInstantID(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "instantid": (data_types.INSTANTID,),
                "insightface": (data_types.FACEANALYSIS,),
                "control_net": (data_types.CONTROL_NET,),
                "image": ("IMAGE",),
                "model": (data_types.MODEL,),
                "positive": (data_types.CONDITIONING,),
                "negative": (data_types.CONDITIONING,),
                "weight": (
                    "FLOAT",
                    {
                        "default": 0.8,
                        "min": 0.0,
                        "max": 5.0,
                        "step": 0.01,
                    },
                ),
                "start_at": (
                    "FLOAT",
                    {
                        "default": 0.0,
                        "min": 0.0,
                        "max": 1.0,
                        "step": 0.001,
                    },
                ),
                "end_at": (
                    "FLOAT",
                    {
                        "default": 1.0,
                        "min": 0.0,
                        "max": 1.0,
                        "step": 0.001,
                    },
                ),
            },
            "optional": {
                "image_kps": ("IMAGE",),
                "mask": ("MASK",),
            },
        }

    RETURN_TYPES = (
        data_types.MODEL,
        data_types.CONDITIONING,
        data_types.CONDITIONING,
    )
    RETURN_NAMES = (
        "MODEL",
        "positive",
        "negative",
    )
    # FUNCTION = "apply_instantid" use default_function
    CATEGORY = "InstantID"
    NODE_DISPLAY_NAME = "Apply InstantID"


class ApplyInstantIDAdvanced(ApplyInstantID):
    NODE_DISPLAY_NAME = "Apply InstantID Adavanced"

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "instantid": (data_types.INSTANTID,),
                "insightface": (data_types.FACEANALYSIS,),
                "control_net": (data_types.CONTROL_NET,),
                "image": ("IMAGE",),
                "model": (data_types.MODEL,),
                "positive": (data_types.CONDITIONING,),
                "negative": (data_types.CONDITIONING,),
                "ip_weight": (
                    "FLOAT",
                    {
                        "default": 0.8,
                        "min": 0.0,
                        "max": 3.0,
                        "step": 0.01,
                    },
                ),
                "cn_strength": (
                    "FLOAT",
                    {
                        "default": 0.8,
                        "min": 0.0,
                        "max": 10.0,
                        "step": 0.01,
                    },
                ),
                "start_at": (
                    "FLOAT",
                    {
                        "default": 0.0,
                        "min": 0.0,
                        "max": 1.0,
                        "step": 0.001,
                    },
                ),
                "end_at": (
                    "FLOAT",
                    {
                        "default": 1.0,
                        "min": 0.0,
                        "max": 1.0,
                        "step": 0.001,
                    },
                ),
                "noise": (
                    "FLOAT",
                    {
                        "default": 0.0,
                        "min": 0.0,
                        "max": 1.0,
                        "step": 0.1,
                    },
                ),
                "combine_embeds": (
                    ["average", "norm average", "concat"],
                    {"default": "average"},
                ),
            },
            "optional": {
                "image_kps": ("IMAGE",),
                "mask": ("MASK",),
            },
        }


class InstantIDFaceAnalysis(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "provider": (["CUDA"],),
            },
        }

    RETURN_TYPES = (data_types.FACEANALYSIS,)
    # FUNCTION = "load_insight_face"
    CATEGORY = "InstantID"
    NODE_DISPLAY_NAME = "InstantID Face Analysis"



import base64
import hashlib
import os
import re
import urllib.request

import folder_paths
from bizyengine.core import BizyAirBaseNode
from bizyengine.core.image_utils import encode_data
from nodes import LoadImage


class LoadImageURL(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "url": (
                    "STRING",
                    {
                        "multiline": True,  # True if you want the field to look like the one on the ClipTextEncode node
                        "default": "url or path",
                        "lazy": True,
                    },
                ),
            },
        }

    CATEGORY = "image_utils"
    RETURN_TYPES = ("IMAGE", "MASK")
    FUNCTION = "apply"
    NODE_DISPLAY_NAME = "Load Image (URL)"

    def apply(self, url: str):
        url = url.strip()
        input_dir = folder_paths.get_input_directory()

        # check if it's a base64 encoded image
        base64_pattern = r"^data:image/([a-zA-Z]+);base64,"
        match = re.match(base64_pattern, url)

        if match:
            # get image format
            image_format = match.group(1)
            image_data = url.split(",")[1]

            # calculate hash of the base64 data
            hash_object = hashlib.md5(image_data.encode())
            hash_value = hash_object.hexdigest()

            filename = f"base64-{hash_value}.{image_format}"
            file_path = os.path.join(input_dir, filename)

            if not os.path.exists(file_path):
                with open(file_path, "wb") as f:
                    f.write(base64.b64decode(image_data))
                print(f"Base64 image saved as {filename}")
            else:
                print(f"Base64 image {filename} already exists, skipping save.")
        else:
            filename = os.path.basename(url)
            file_path = os.path.join(input_dir, filename)

            if os.path.exists(file_path):
                print(f"File {filename} already exists, skipping download.")
            else:
                urllib.request.urlretrieve(url, file_path)
                print(f"Image successfully downloaded and saved as {filename}.")

        return LoadImage().load_image(filename)


class Image_Encode(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {"image": ("IMAGE",)},
            "optional": {
                "lossless": (
                    "BOOLEAN",
                    {"default": False, "label_on": "yes", "label_off": "no"},
                ),
            },
        }

    CATEGORY = "image_utils"
    FUNCTION = "apply"
    RETURN_TYPES = ("IMAGE",)
    NODE_DISPLAY_NAME = "Image Encode"

    def apply(self, image, lossless=False):
        out = encode_data(image, lossless=lossless)
        return (out,)



import json
import os
import pprint
import urllib.error
import urllib.request
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Any, Union

import aiohttp
import comfy

from .caching import CacheManager

__all__ = ["send_request"]

from dataclasses import dataclass, field

from .env_var import BIZYAIR_API_KEY, BIZYAIR_DEBUG, BIZYAIR_SERVER_ADDRESS

IS_API_KEY_VALID = None


@dataclass
class APIKeyState:
    current_api_key: str = field(default=None)
    is_valid: bool = field(default=None)


api_key_state = APIKeyState()


def set_api_key(api_key: str = "YOUR_API_KEY", override: bool = False):
    global BIZYAIR_API_KEY, api_key_state
    if api_key_state.is_valid is not None and not override:
        warnings.warn("API key has already been set and will not be overridden.")
        return
    if validate_api_key(api_key):
        BIZYAIR_API_KEY = api_key
        api_key_state.is_valid = True
        print("\033[92mAPI key is set successfully.\033[0m")
    else:
        api_key_state.is_valid = False
        warnings.warn("Invalid API key provided.")


def validate_api_key(api_key: str = None) -> bool:
    global api_key_state
    if not api_key or not isinstance(api_key, str):
        warnings.warn("API key is not set.")
        return False
    # if api_key_state.current_api_key == api_key and api_key_state.is_valid is not None:
    #     return api_key_state.is_valid
    api_key_state.current_api_key = api_key
    url = f"{BIZYAIR_SERVER_ADDRESS}/user/info"
    headers = {"accept": "application/json", "authorization": f"Bearer {api_key}"}
    try:
        response_data = send_request(
            method="GET", url=url, headers=headers, callback=None
        )
        if "message" not in response_data or response_data["message"] != "Ok":
            api_key_state.is_valid = False
            raise ValueError(
                f"\033[91mAPI key validation failed. API Key: {api_key}\033[0m"
            )
        else:
            api_key_state.is_valid = True
    except ConnectionError as ce:
        api_key_state.is_valid = False
        raise ValueError(f"\033[91mConnection error: {ce}\033[0m")
    except PermissionError as pe:
        api_key_state.is_valid = False
        raise ValueError(
            f"\033[91mError validating API key: {api_key}, error: {pe}\033[0m"
        )
    except Exception as e:
        api_key_state.is_valid = False
        raise ValueError(f"\033[91mOther error: {e}\033[0m")
    return api_key_state.is_valid


def get_api_key() -> str:
    global BIZYAIR_API_KEY
    try:
        validate_api_key(BIZYAIR_API_KEY)
    except Exception as e:
        print(str(e))
        raise ValueError(str(e))
    return BIZYAIR_API_KEY


def _headers():
    headers = {
        "accept": "application/json",
        "content-type": "application/json",
        "authorization": f"Bearer {get_api_key()}",
    }
    return headers


def process_response_data(response_data: dict) -> dict:
    # Check if 'result' key exists, indicating a cloud response
    if "result" in response_data:
        try:
            msg = json.loads(response_data["result"])
        except json.JSONDecodeError:
            raise ValueError(f"Failed to decode JSON from response. {response_data=}")
    else:
        # Handle local response directly
        msg = response_data

    return msg  # Return processed data or modify as needed


def send_request(
    method: str = "POST",
    url: str = None,
    data: bytes = None,
    verbose=False,
    callback: callable = process_response_data,
    response_handler: callable = json.loads,
    cache_manager: CacheManager = None,
    **kwargs,
) -> Union[dict, Any]:
    try:
        headers = kwargs.pop("headers") if "headers" in kwargs else _headers()
        headers["User-Agent"] = "BizyAir Client"

        req = urllib.request.Request(
            url, data=data, headers=headers, method=method, **kwargs
        )
        with urllib.request.urlopen(req) as response:
            response_data = response.read().decode("utf-8")
    except urllib.error.URLError as e:
        error_message = str(e)
        response_body = e.read().decode("utf-8") if hasattr(e, "read") else "N/A"
        if verbose:
            print(f"URLError encountered: {error_message}")
            print(f"Response Body: {response_data}")
        code, message = "N/A", "N/A"
        try:
            response_dict = json.loads(response_body)
            if isinstance(response_dict, dict):
                code = response_dict.get("code", "N/A")
                message = response_dict.get("message", "N/A")

        except json.JSONDecodeError:
            if verbose:
                print("Failed to decode response body as JSON.")

        if "Unauthorized" in error_message:
            raise PermissionError(
                "Key is invalid, please refer to https://cloud.siliconflow.cn to get the API key.\n"
                "If you have the key, please click the 'API Key' button at the bottom right to set the key."
            )
        elif code != "N/A" and message != "N/A":
            raise ConnectionError(
                f"Failed to handle your request: {error_message}.\n"
                + f"    Error code: {code}.\n"
                + f"    Error message: {message}\n."
                + "The cause of this issue may be incorrect parameter status or ongoing background tasks. \n"
                + "If retrying after waiting for a while still does not resolve the issue, please report it to "
                + "Bizyair's official support."
            )
        else:
            common_sites = [
                "https://www.baidu.com",
                "https://www.bing.com",
                "https://www.alibaba.com",
            ]
            results = {}
            for site in common_sites:
                success = False
                try:

                    class NoRedirectHandler(urllib.request.HTTPRedirectHandler):
                        def redirect_request(self, req, fp, code, msg, headers, newurl):
                            return None

                    opener = urllib.request.build_opener(NoRedirectHandler())
                    response = opener.open(site, timeout=5)
                    success = 200 <= response.getcode() < 400
                except urllib.error.HTTPError as e:
                    success = 200 <= e.code < 400
                except (urllib.error.URLError, TimeoutError):
                    pass
                results[site] = "Success" if success else "Failed"
            raise ConnectionError(
                f"Failed to connect to the server: {url}.\n"
                + "The connection attempts to the public sites return the following results:\n"
                + "\n".join(
                    [f"    {site}: {result}" for site, result in results.items()]
                )
                + "\nPlease check the network connection."
            )

    if response_handler:
        response_data = response_handler(response_data)
    if callback:
        return callback(response_data)
    return response_data


async def async_send_request(
    method: str = "POST",
    url: str = None,
    data: bytes = None,
    verbose=False,
    callback: callable = process_response_data,
    **kwargs,
) -> dict:
    headers = kwargs.pop("headers") if "headers" in kwargs else _headers()
    try:
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method, url, data=data, headers=headers, **kwargs
            ) as response:
                response_data = await response.text()
                if response.status != 200:
                    error_message = f"HTTP Status {response.status}"
                    if verbose:
                        print(f"Error encountered: {error_message}")
                    if response.status == 401:
                        raise PermissionError(
                            "Key is invalid, please refer to https://cloud.siliconflow.cn to get the API key.\n"
                            "If you have the key, please click the 'BizyAir Key' button at the bottom right to set the key."
                        )
                    else:
                        raise ConnectionError(
                            f"Failed to connect to the server: {error_message}.\n"
                            + "Please check your API key and ensure the server is reachable.\n"
                            + "Also, verify your network settings and disable any proxies if necessary.\n"
                            + "After checking, please restart the ComfyUI service."
                        )
                if callback:
                    return callback(json.loads(response_data))
                return json.loads(response_data)
    except aiohttp.ClientError as e:
        print(f"Error fetching data: {e}")
        return {}
    except Exception as e:
        print(f"Error fetching data: {str(e)}")
        return {}


def fetch_models_by_type(
    url: str, model_type: str, *, method="GET", verbose=False
) -> dict:
    if not validate_api_key(BIZYAIR_API_KEY):
        return {}

    payload = {"type": model_type}
    if BIZYAIR_DEBUG:
        pprint.pprint(payload)
    msg = send_request(
        method=method,
        url=url,
        data=json.dumps(payload).encode("utf-8"),
        verbose=verbose,
    )
    return msg



import json
from collections import deque
from typing import Any, Dict, List

from bizyengine.core.commands.base import Processor  # type: ignore
from bizyengine.core.common import client, get_api_key
from bizyengine.core.common.caching import BizyAirTaskCache, CacheConfig
from bizyengine.core.common.env_var import (
    BIZYAIR_DEBUG,
    BIZYAIR_DEV_REQUEST_URL,
    BIZYAIR_SERVER_ADDRESS,
)
from bizyengine.core.configs.conf import ModelRule, config_manager
from bizyengine.core.path_utils import (
    convert_prompt_label_path_to_real_path,
    guess_url_from_node,
)


def is_link(obj):
    if not isinstance(obj, list):
        return False
    if len(obj) != 2:
        return False
    if not isinstance(obj[0], str):
        return False
    if not isinstance(obj[1], int) and not isinstance(obj[1], float):
        return False
    return True


from dataclasses import dataclass


class SearchServiceRouter(Processor):
    def process(self, prompt: Dict[str, Dict[str, Any]], last_node_ids: List[str]):
        if BIZYAIR_DEV_REQUEST_URL:
            return BIZYAIR_DEV_REQUEST_URL

        # TODO Improve distribution logic
        queue = deque(last_node_ids)
        visited = {key: True for key in last_node_ids}
        results: List[ModelRule] = []
        class_type_table = {
            node_data["class_type"]: True for node_data in prompt.values()
        }

        while queue:
            vertex = queue.popleft()
            if BIZYAIR_DEBUG:
                print(vertex, end="->")

            rules = guess_url_from_node(prompt[vertex], class_type_table)
            if rules:
                results.extend(rules)
            for _, in_data in prompt[vertex].get("inputs", {}).items():
                if is_link(in_data):
                    neighbor = in_data[0]
                    if neighbor not in visited:
                        visited[neighbor] = True
                        queue.append(neighbor)

        base_model, out_route, out_score = None, None, 0
        for rule in results[::-1]:
            # TODO add to config models.yaml
            if rule.mode_type in {"unet", "vae", "checkpoint", "upscale_models"}:
                base_model = rule.base_model
                out_route = rule.route
                out_score = rule.score
                break

        for rule in results:
            if base_model is None:
                if rule.score > out_score:
                    out_route, out_score = rule.route, rule.score
            if rule.base_model == base_model:
                if rule.score > out_score:
                    out_route, out_score = rule.route, rule.score
        assert (
            out_route is not None
        ), "Failed to find out_route, please check your prompt"
        return f"{BIZYAIR_SERVER_ADDRESS}{out_route}"

    def validate_input(
        self, prompt: Dict[str, Dict[str, Any]], last_node_ids: List[str]
    ):
        assert len(last_node_ids) == 1
        return True


class PromptProcessor(Processor):
    def _exec_info(self, prompt: Dict[str, Dict[str, Any]]):
        exec_info = {
            "model_version_ids": [],
            "api_key": get_api_key(),
        }

        model_version_id_prefix = config_manager.get_model_version_id_prefix()
        for node_id, node_data in prompt.items():
            for k, v in node_data.get("inputs", {}).items():
                if isinstance(v, str) and v.startswith(model_version_id_prefix):
                    model_version_id = int(v[len(model_version_id_prefix) :])
                    exec_info["model_version_ids"].append(model_version_id)
        return exec_info

    def process(
        self, url: str, prompt: Dict[str, Dict[str, Any]], last_node_ids: List[str]
    ):
        return client.send_request(
            url=url,
            data=json.dumps(
                {
                    "prompt": prompt,
                    "last_node_id": last_node_ids[0],
                    "exec_info": self._exec_info(prompt),
                }
            ).encode("utf-8"),
        )

    def validate_input(
        self, url: str, prompt: Dict[str, Dict[str, Any]], last_node_ids: List[str]
    ):
        return True



import os

from bizyengine.core import BizyAirBaseNode, BizyAirNodeIO, create_node_data
from bizyengine.core import path_utils as folder_paths
from bizyengine.core.data_types import CLIP, CONDITIONING, CONTROL_NET, MODEL

AUTHOR_NAME = "MinusZone"
CATEGORY_NAME = f"Kolors"


class MZ_KolorsUNETLoaderV2(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "unet_name": (folder_paths.get_filename_list("unet"),),
            }
        }

    RETURN_TYPES = (MODEL,)
    RETURN_NAMES = ("model",)

    FUNCTION = "load_unet"

    CATEGORY = CATEGORY_NAME
    NODE_DISPLAY_NAME = f"{AUTHOR_NAME} - KolorsUNETLoaderV2"

    def load_unet(self, **kwargs):

        node_data = create_node_data(
            class_type="MZ_KolorsUNETLoaderV2",
            inputs=kwargs,
            outputs={"slot_index": 0},
        )
        config_file = folder_paths.guess_config(unet_name=kwargs["unet_name"])
        out = BizyAirNodeIO(
            self.assigned_id, {self.assigned_id: node_data}, config_file=config_file
        )
        return (out,)


WEIGHT_TYPES = [
    "linear",
    "ease in",
    "ease out",
    "ease in-out",
    "reverse in-out",
    "weak input",
    "weak output",
    "weak middle",
    "strong middle",
    "style transfer",
    "composition",
    "strong style transfer",
    "style and composition",
    "style transfer precise",
    "composition precise",
]


class MZ_KolorsControlNetLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "control_net_name": (folder_paths.get_filename_list("controlnet"),),
                # "seed": ("INT", {"default": 0, "min": 0, "max": 1000000}),
            }
        }

    RETURN_TYPES = (CONTROL_NET,)
    RETURN_NAMES = ("ControlNet",)
    FUNCTION = "load_controlnet"

    CATEGORY = CATEGORY_NAME
    NODE_DISPLAY_NAME = f"{AUTHOR_NAME} - KolorsControlNetLoader"

    def load_controlnet(self, **kwargs):
        node_data = create_node_data(
            class_type="MZ_KolorsControlNetLoader",
            inputs=kwargs,
            outputs={"slot_index": 0},
        )
        assigned_id = self.assigned_id
        node = BizyAirNodeIO(assigned_id, {assigned_id: node_data})
        return (node,)



import os
import uuid

import torch
from bizyengine.core import BizyAirBaseNode, BizyAirNodeIO, create_node_data
from bizyengine.core.common.env_var import BIZYAIR_SERVER_ADDRESS
from bizyengine.core.data_types import CONDITIONING
from bizyengine.core.image_utils import encode_data

from .utils import (
    decode_and_deserialize,
    get_api_key,
    send_post_request,
    serialize_and_encode,
)

CATEGORY_NAME = "☁️BizyAir/Kolors"


class BizyAirMZChatGLM3TextEncode:
    API_URL = f"{BIZYAIR_SERVER_ADDRESS}/supernode/mzkolorschatglm3"

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "text": ("STRING", {"multiline": True, "dynamicPrompts": True}),
            }
        }

    RETURN_TYPES = ("CONDITIONING",)

    FUNCTION = "encode"
    CATEGORY = CATEGORY_NAME

    def encode(self, text):
        API_KEY = get_api_key()
        assert len(text) <= 4096, f"the prompt is too long, length: {len(text)}"

        payload = {
            "text": text,
        }
        auth = f"Bearer {API_KEY}"
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "authorization": auth,
        }

        response: str = send_post_request(
            self.API_URL, payload=payload, headers=headers
        )
        tensors_np = decode_and_deserialize(response)

        ret_conditioning = []
        for item in tensors_np:
            t, d = item
            t_tensor = torch.from_numpy(t)
            d_dict = {}
            for k, v in d.items():
                d_dict[k] = torch.from_numpy(v)
            ret_conditioning.append([t_tensor, d_dict])

        return (ret_conditioning,)


class BizyAir_MinusZoneChatGLM3TextEncode(BizyAirMZChatGLM3TextEncode, BizyAirBaseNode):
    RETURN_TYPES = (CONDITIONING,)

    FUNCTION = "mz_encode"

    def mz_encode(self, text):
        out = self.encode(text)[0]
        node_data = create_node_data(
            class_type="ComfyAirLoadData",
            inputs={"conditioning": {"relay": out}},
            outputs={"slot_index": 3},
        )
        node_data["is_changed"] = uuid.uuid4().hex
        return (
            BizyAirNodeIO(
                self.assigned_id,
                nodes={self.assigned_id: encode_data(node_data, old_version=True)},
            ),
        )


NODE_CLASS_MAPPINGS = {
    "BizyAir_MinusZoneChatGLM3TextEncode": BizyAir_MinusZoneChatGLM3TextEncode,
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "BizyAir_MinusZoneChatGLM3TextEncode": "☁️BizyAir MinusZone ChatGLM3 Text Encode",
}



from bizyengine.core import BizyAirBaseNode, data_types


class DisableNoise(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {"required": {}}

    RETURN_TYPES = ("NOISE",)
    CATEGORY = "sampling/custom_sampling/noise"


class VPScheduler(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "steps": ("INT", {"default": 20, "min": 1, "max": 10000}),
                "beta_d": (
                    "FLOAT",
                    {
                        "default": 19.9,
                        "min": 0.0,
                        "max": 5000.0,
                        "step": 0.01,
                        "round": False,
                    },
                ),  # TODO: fix default values
                "beta_min": (
                    "FLOAT",
                    {
                        "default": 0.1,
                        "min": 0.0,
                        "max": 5000.0,
                        "step": 0.01,
                        "round": False,
                    },
                ),
                "eps_s": (
                    "FLOAT",
                    {
                        "default": 0.001,
                        "min": 0.0,
                        "max": 1.0,
                        "step": 0.0001,
                        "round": False,
                    },
                ),
            }
        }

    RETURN_TYPES = ("SIGMAS",)
    CATEGORY = "sampling/custom_sampling/schedulers"


class SplitSigmas(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "sigmas": ("SIGMAS",),
                "step": ("INT", {"default": 0, "min": 0, "max": 10000}),
            }
        }

    RETURN_TYPES = ("SIGMAS", "SIGMAS")
    RETURN_NAMES = ("high_sigmas", "low_sigmas")
    CATEGORY = "sampling/custom_sampling/sigmas"


class SplitSigmasDenoise(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "sigmas": ("SIGMAS",),
                "denoise": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
            }
        }

    RETURN_TYPES = ("SIGMAS", "SIGMAS")
    RETURN_NAMES = ("high_sigmas", "low_sigmas")
    CATEGORY = "sampling/custom_sampling/sigmas"

    # FUNCTION = "get_sigmas"


class FlipSigmas(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "sigmas": ("SIGMAS",),
            }
        }

    RETURN_TYPES = ("SIGMAS",)
    CATEGORY = "sampling/custom_sampling/sigmas"

    # FUNCTION = "get_sigmas"


class CFGGuider(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (data_types.MODEL,),
                "positive": (data_types.CONDITIONING,),
                "negative": (data_types.CONDITIONING,),
                "cfg": (
                    "FLOAT",
                    {
                        "default": 8.0,
                        "min": 0.0,
                        "max": 100.0,
                        "step": 0.1,
                        "round": 0.01,
                    },
                ),
            }
        }

    RETURN_TYPES = ("GUIDER",)

    # FUNCTION = "get_guider"
    CATEGORY = "sampling/custom_sampling/guiders"



from bizyengine.core import BizyAirBaseNode, data_types


class InstructPixToPixConditioning(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "positive": (data_types.CONDITIONING,),
                "negative": (data_types.CONDITIONING,),
                "vae": (data_types.VAE,),
                "pixels": ("IMAGE",),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING, data_types.CONDITIONING, "LATENT")
    RETURN_NAMES = ("positive", "negative", "latent")
    # FUNCTION = "encode"

    CATEGORY = "conditioning/instructpix2pix"



from bizyengine.core import BizyAirBaseNode

# layerstyle advance
NODE_NAME = "SegmentAnythingUltra V2"
sam_model_dir_name = "sams"
sam_model_list = {
    "sam_vit_h (2.56GB)": {
        "model_url": "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_h_4b8939.pth"
    },
    # "sam_vit_l (1.25GB)": {
    #     "model_url": "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_l_0b3195.pth"
    # },
    # "sam_vit_b (375MB)": {
    #     "model_url": "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_b_01ec64.pth"
    # },
    # "sam_hq_vit_h (2.57GB)": {
    #     "model_url": "https://huggingface.co/lkeab/hq-sam/resolve/main/sam_hq_vit_h.pth"
    # },
    # "sam_hq_vit_l (1.25GB)": {
    #     "model_url": "https://huggingface.co/lkeab/hq-sam/resolve/main/sam_hq_vit_l.pth"
    # },
    # "sam_hq_vit_b (379MB)": {
    #     "model_url": "https://huggingface.co/lkeab/hq-sam/resolve/main/sam_hq_vit_b.pth"
    # },
    # "mobile_sam(39MB)": {
    #     "model_url": "https://github.com/ChaoningZhang/MobileSAM/blob/master/weights/mobile_sam.pt"
    # }
}

groundingdino_model_dir_name = "grounding-dino"
groundingdino_model_list = {
    "GroundingDINO_SwinT_OGC (694MB)": {
        "config_url": "https://huggingface.co/ShilongLiu/GroundingDINO/resolve/main/GroundingDINO_SwinT_OGC.cfg.py",
        "model_url": "https://huggingface.co/ShilongLiu/GroundingDINO/resolve/main/groundingdino_swint_ogc.pth",
    },
    # "GroundingDINO_SwinB (938MB)": {
    #     "config_url": "https://huggingface.co/ShilongLiu/GroundingDINO/resolve/main/GroundingDINO_SwinB.cfg.py",
    #     "model_url": "https://huggingface.co/ShilongLiu/GroundingDINO/resolve/main/groundingdino_swinb_cogcoor.pth"
    # },
}


def list_sam_model():
    return list(sam_model_list.keys())


def list_groundingdino_model():
    return list(groundingdino_model_list.keys())


class SegmentAnythingUltraV2(BizyAirBaseNode):

    CLASS_TYPE_NAME = "LayerMask: SegmentAnythingUltra V2"
    NODE_DISPLAY_NAME = "LayerMask: SegmentAnythingUltra V2(Advance)"

    def __init__(self):
        self.SAM_MODEL = None
        self.DINO_MODEL = None
        self.previous_sam_model = ""
        self.previous_dino_model = ""
        pass

    @classmethod
    def INPUT_TYPES(cls):

        method_list = [
            "VITMatte",
            "VITMatte(local)",
            "PyMatting",
            "GuidedFilter",
        ]
        device_list = ["cuda"]
        return {
            "required": {
                "image": ("IMAGE",),
                "sam_model": (list_sam_model(),),
                "grounding_dino_model": (list_groundingdino_model(),),
                "threshold": (
                    "FLOAT",
                    {"default": 0.3, "min": 0, "max": 1.0, "step": 0.01},
                ),
                "detail_method": (method_list,),
                "detail_erode": (
                    "INT",
                    {"default": 6, "min": 1, "max": 255, "step": 1},
                ),
                "detail_dilate": (
                    "INT",
                    {"default": 6, "min": 1, "max": 255, "step": 1},
                ),
                "black_point": (
                    "FLOAT",
                    {
                        "default": 0.15,
                        "min": 0.01,
                        "max": 0.98,
                        "step": 0.01,
                        "display": "slider",
                    },
                ),
                "white_point": (
                    "FLOAT",
                    {
                        "default": 0.99,
                        "min": 0.02,
                        "max": 0.99,
                        "step": 0.01,
                        "display": "slider",
                    },
                ),
                "process_detail": ("BOOLEAN", {"default": True}),
                "prompt": ("STRING", {"default": "subject"}),
                "device": (device_list,),
                "max_megapixels": (
                    "FLOAT",
                    {"default": 2.0, "min": 1, "max": 999, "step": 0.1},
                ),
                "cache_model": ("BOOLEAN", {"default": True}),
            },
            "optional": {},
        }

    RETURN_TYPES = (
        "IMAGE",
        "MASK",
    )
    RETURN_NAMES = (
        "image",
        "mask",
    )
    # FUNCTION = "segment_anything_ultra_v2"
    CATEGORY = "😺dzNodes/LayerMask"


NODE_CLASS_MAPPINGS = {
    "LayerMask: SegmentAnythingUltra V2": SegmentAnythingUltraV2,
}

NODE_DISPLAY_NAME_MAPPINGS = {
    "LayerMask: SegmentAnythingUltra V2": "LayerMask: SegmentAnythingUltra V2(Advance)",
}



"""
pip install numpy scikit-image pillow

"""

import os
import random

import folder_paths
import numpy as np
from bizyengine.core import NODE_CLASS_MAPPINGS
from bizyengine.core.image_utils import decode_data, encode_data
from PIL import Image


class ImagesTest:
    # https://docs.comfy.org/essentials/custom_node_images_and_masks#images
    def __init__(self):
        self.output_dir = folder_paths.get_temp_directory()
        self.type = "temp"
        self.prefix_append = "_temp_" + "".join(
            random.choice("abcdefghijklmnopqrstupvxyz") for x in range(5)
        )

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "images1": ("IMAGE",),
                "images2": ("IMAGE",),
                "ssim_threshold": ("STRING", {"default": "0.9"}),
                "raise_if_diff": (["enable", "disable"],),
                "image_id": ("STRING", {"default": "ComfyUI"}),
            },
            "hidden": {"prompt": "PROMPT", "extra_pnginfo": "EXTRA_PNGINFO"},
        }

    RETURN_TYPES = ()
    FUNCTION = "save_images"

    OUTPUT_NODE = True

    CATEGORY = "images_test"

    def save_images(
        self,
        images1,
        images2,
        ssim_threshold: float,
        raise_if_diff: str,
        image_id="ComfyUI",
        prompt=None,
        extra_pnginfo=None,
    ):
        from skimage.metrics import structural_similarity as ssim

        assert len(images1) == len(images2)
        filename_prefix = image_id[:]
        filename_prefix += self.prefix_append
        (
            full_output_folder,
            filename,
            counter,
            subfolder,
            filename_prefix,
        ) = folder_paths.get_save_image_path(
            filename_prefix,
            self.output_dir,
            images1[0].shape[1],
            images1[0].shape[0],
        )
        results = list()
        for image1, image2 in zip(images1, images2):

            # image diff
            image = image1.cpu() - image2.cpu()

            i = 255.0 * image.cpu().numpy()
            img = Image.fromarray(np.clip(i, 0, 255).astype(np.uint8))
            metadata = None
            file = f"{filename}_{counter:05}_.png"
            img.save(
                os.path.join(full_output_folder, file),
                pnginfo=metadata,
                compress_level=4,
            )
            results.append(
                {"filename": file, "subfolder": subfolder, "type": self.type}
            )
            counter += 1

            max_diff = image.abs().max().item()

            img1 = self.image_to_numpy(image1)
            img2 = self.image_to_numpy(image2)
            ssim = ssim(img1, img2, channel_axis=2)

            print(
                "\033[91m"
                f"[ShowImageDiff {image_id}] Max value of diff is {max_diff}, image simularity is {ssim:.6f}"
                + "\033[0m"
            )

            if raise_if_diff == "enable":
                assert ssim > float(
                    ssim_threshold
                ), f"Image diff is too large, ssim is {ssim}, "

        return {"ui": {"images": results}}

    def image_to_numpy(self, image):
        i = 255.0 * image.cpu().numpy()
        img = Image.fromarray(np.clip(i, 0, 255).astype(np.uint8))
        return np.array(img)


class ImageEncodeDecodeTest:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {"image": ("IMAGE",)},
            "optional": {
                "lossless": (
                    "BOOLEAN",
                    {"default": False, "label_on": "yes", "label_off": "no"},
                ),
            },
        }

    CATEGORY = "images_test"
    FUNCTION = "test"
    RETURN_TYPES = ("IMAGE",)

    def test(self, image, lossless=False):
        return (decode_data(encode_data(image, lossless=lossless)),)


NODE_CLASS_MAPPINGS["Tools_ImagesTest"] = ImagesTest
NODE_CLASS_MAPPINGS["Tools_ImageEncodeDecodeTest"] = ImageEncodeDecodeTest



from bizyengine.core import BizyAirBaseNode


class Wan_Model_Loader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "ckpt_name": (("Wan2.1-T2V-1.3B",),),
            }
        }

    RETURN_TYPES = ("WAN_MODEL",)
    RETURN_NAMES = ("model",)
    CATEGORY = "WanT2V"


class Wan_T2V_Pipeline(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "model": ("WAN_MODEL",),
                "prompt": (
                    "STRING",
                    {
                        "multiline": True,
                        "default": "Two anthropomorphic cats in comfy boxing gear and bright gloves fight intensely on a spotlighted stage.",
                    },
                ),
                "resolution": (
                    ["480*832", "832*480", "624*624", "704*544", "544*704"],
                    {"default": "480*832"},
                ),
                "sampling_steps": ("INT", {"default": 50, "min": 1, "max": 50}),
                "guidance_scale": ("FLOAT", {"default": 6.0, "min": 0, "max": 20}),
                "shift_scale": ("FLOAT", {"default": 8.0, "min": 0, "max": 20}),
                "seed": ("INT", {"default": -1, "min": -1, "max": 0xFFFFFFFFFFFFFFFF}),
                "negative_prompt": (
                    "STRING",
                    {"multiline": True, "default": "Low quality, blurry"},
                ),
            }
        }

    RETURN_TYPES = ("STRING",)
    RETURN_NAMES = ("video_url",)
    CATEGORY = "WanT2V"
    # FUNCTION = "generate_video"



import hashlib
import json
import os
import uuid
from enum import Enum

import folder_paths
import node_helpers
import numpy as np
import torch
from bizyengine.core.common.env_var import BIZYAIR_SERVER_ADDRESS
from bizyengine.core.image_utils import (
    decode_base64_to_np,
    decode_data,
    encode_data,
    encode_image_to_base64,
)
from nodes import LoadImage
from PIL import Image, ImageOps, ImageSequence

from .utils import (
    decode_and_deserialize,
    get_api_key,
    send_post_request,
    serialize_and_encode,
)


class RemoveBackground:
    API_URL = f"{BIZYAIR_SERVER_ADDRESS}/supernode/removebg"

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "image": ("IMAGE",),
            }
        }

    RETURN_TYPES = ("IMAGE", "MASK")
    FUNCTION = "remove_background"

    CATEGORY = "☁️BizyAir"

    def remove_background(self, image):
        API_KEY = get_api_key()
        device = image.device
        _, h, w, _ = image.shape
        assert (
            w <= 1536 and h <= 1536
        ), f"width and height must be less than 1536, but got {w} and {h}"

        payload = {
            "is_compress": True,
            "image": None,
        }
        auth = f"Bearer {API_KEY}"
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "authorization": auth,
        }
        input_image, compress = serialize_and_encode(image, compress=True)
        payload["image"] = input_image
        payload["is_compress"] = compress

        response: str = send_post_request(
            self.API_URL, payload=payload, headers=headers
        )
        tensors = decode_and_deserialize(response)
        t_images = tensors["images"].to(device)
        t_mask = tensors["mask"].to(device)
        return (t_images, t_mask)


class GenerateLightningImage:
    API_URL = f"{BIZYAIR_SERVER_ADDRESS}/supernode/realvis4lightning"

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "prompt": (
                    "STRING",
                    {"multiline": True, "dynamicPrompts": True, "default": "a dog"},
                ),
                "seed": ("INT", {"default": 1, "min": 0, "max": 0xFFFFFFFFFFFFFFFF}),
                "width": ("INT", {"default": 1024, "min": 16, "max": 1024, "step": 8}),
                "height": ("INT", {"default": 1024, "min": 16, "max": 1024, "step": 8}),
                "cfg": (
                    "FLOAT",
                    {
                        "default": 1.5,
                        "min": 0.0,
                        "max": 10.0,
                        "step": 0.1,
                        "round": 0.01,
                    },
                ),
                "batch_size": ("INT", {"default": 1, "min": 1, "max": 4}),
            }
        }

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "generate_image"

    CATEGORY = "☁️BizyAir"

    def generate_image(self, prompt, seed, width, height, cfg, batch_size):
        API_KEY = get_api_key()
        assert (
            width <= 1024 and height <= 1024
        ), f"width and height must be less than 1024, but got {width} and {height}"

        payload = {
            "batch_size": batch_size,
            "width": width,
            "height": height,
            "prompt": prompt,
            "cfg": cfg,
            "seed": seed,
        }
        auth = f"Bearer {API_KEY}"
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "authorization": auth,
        }

        response: str = send_post_request(
            self.API_URL, payload=payload, headers=headers
        )
        tensors_np = decode_and_deserialize(response)
        tensors = torch.from_numpy(tensors_np)

        return (tensors,)


class ToggleServerEndpoint:
    BIZYAIR_SERVER_ENDPOINTS = [
        "https://bizyair-api-st.siliconflow.cn/x/v1",
        "https://bizyair-api.siliconflow.cn/x/v1",
    ]

    def __init__(self):
        self.current_index = 0

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "endpoint": (s.BIZYAIR_SERVER_ENDPOINTS,),
            }
        }

    RETURN_TYPES = ()
    FUNCTION = "toggle_endpoint"
    OUTPUT_NODE = True

    CATEGORY = "☁️BizyAir"

    def toggle_endpoint(self, endpoint):
        BIZYAIR_SERVER_ADDRESS.address = endpoint
        from server import PromptServer

        PromptServer.instance.send_sync(
            "bizyair.server.endpoint.switch",
            {"message": f"Switch server endpoint to {endpoint}"},
        )
        return ()


NODE_CLASS_MAPPINGS = {
    "BizyAirRemoveBackground": RemoveBackground,
    "BizyAirGenerateLightningImage": GenerateLightningImage,
    "BizyAirToggleServerEndpoint": ToggleServerEndpoint,
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "BizyAirRemoveBackground": "☁️BizyAir Remove Image Background",
    "BizyAirGenerateLightningImage": "☁️BizyAir Generate Photorealistic Images",
    "BizyAirToggleServerEndpoint": "☁️BizyAir Switch Server Endpoint",
}



import hashlib
import json
import pprint
import time
import traceback
from dataclasses import dataclass, field
from typing import Any, Dict, List

import comfy
from bizyengine.core.commands.base import Command, Processor  # type: ignore
from bizyengine.core.common.caching import BizyAirTaskCache, CacheConfig
from bizyengine.core.common.client import send_request
from bizyengine.core.common.env_var import (
    BIZYAIR_DEBUG,
    BIZYAIR_DEV_GET_TASK_RESULT_SERVER,
    BIZYAIR_SERVER_ADDRESS,
)
from bizyengine.core.common.utils import truncate_long_strings
from bizyengine.core.configs.conf import config_manager
from bizyengine.core.image_utils import decode_data, encode_data


def get_task_result(task_id: str, offset: int = 0) -> dict:
    """
    Get the result of a task.
    """
    import requests

    task_api = config_manager.get_task_api()
    if BIZYAIR_DEV_GET_TASK_RESULT_SERVER:
        url = f"{BIZYAIR_DEV_GET_TASK_RESULT_SERVER}{task_api.task_result_endpoint}/{task_id}"
    else:
        url = f"{BIZYAIR_SERVER_ADDRESS}{task_api.task_result_endpoint}/{task_id}"

    if BIZYAIR_DEBUG:
        print(f"Debug: get task result url: {url}")
    response_json = send_request(
        method="GET", url=url, data=json.dumps({"offset": offset}).encode("utf-8")
    )
    out = response_json
    events = out.get("data", {}).get("events", [])
    new_events = []
    for event in events:
        if (
            "data" in event
            and isinstance(event["data"], str)
            and event["data"].startswith("https://")
        ):
            # event["data"] = requests.get(event["data"]).json()
            event["data"] = send_request(method="GET", url=event["data"])
        new_events.append(event)
    out["data"]["events"] = new_events
    return out


@dataclass
class BizyAirTask:
    TASK_DATA_STATUS = ["PENDING", "PROCESSING", "COMPLETED"]
    task_id: str
    data_pool: list[dict] = field(default_factory=list)
    data_status: str = None

    @staticmethod
    def check_inputs(inputs: dict) -> bool:
        return (
            inputs.get("code") == 20000
            and inputs.get("status", False)
            and "task_id" in inputs.get("data", {})
        )

    @classmethod
    def from_data(cls, inputs: dict, check_inputs: bool = True) -> "BizyAirTask":
        if check_inputs and not cls.check_inputs(inputs):
            raise ValueError(f"Invalid inputs: {inputs}")
        data = inputs.get("data", {})
        task_id = data.get("task_id", "")
        return cls(task_id=task_id, data_pool=[], data_status="started")

    def is_finished(self) -> bool:
        if not self.data_pool:
            return False
        if self.data_pool[-1].get("data_status") == self.TASK_DATA_STATUS[-1]:
            return True
        return False

    def send_request(self, offset: int = 0) -> dict:
        if offset >= len(self.data_pool):
            return get_task_result(self.task_id, offset)
        else:
            return self.data_pool[offset]

    def get_data(self, offset: int = 0) -> dict:
        if offset >= len(self.data_pool):
            return {}
        return self.data_pool[offset]

    @staticmethod
    def _fetch_remote_data(url: str) -> dict:
        import requests

        return requests.get(url).json()

    def get_last_data(self) -> dict:
        return self.get_data(len(self.data_pool) - 1)

    def do_task_until_completed(
        self, *, timeout: int = 600, poll_interval: float = 1
    ) -> list[dict]:
        offset = 0
        start_time = time.time()
        pbar = None
        while not self.is_finished():
            try:
                data = self.send_request(offset)
                data_lst = self._extract_data_list(data)
                self.data_pool.extend(data_lst)
                offset += len(data_lst)
                for data in data_lst:
                    message = data.get("data", {}).get("message", {})
                    if (
                        isinstance(message, dict)
                        and message.get("event", None) == "progress"
                    ):
                        value = message["data"]["value"]
                        total = message["data"]["max"]
                        if pbar is None:
                            pbar = comfy.utils.ProgressBar(total)
                        pbar.update_absolute(value + 1, total, None)
            except Exception as e:
                print(f"Exception: {e}")

            if time.time() - start_time > timeout:
                raise TimeoutError(f"Timeout waiting for task {self.task_id} to finish")

            time.sleep(poll_interval)

        return self.data_pool

    def _extract_data_list(self, data):
        data_lst = data.get("data", {}).get("events", [])
        if not data_lst:
            raise ValueError(f"No data found in task {self.task_id}")
        return data_lst


class PromptServer(Command):
    cache_manager: BizyAirTaskCache = BizyAirTaskCache(
        config=CacheConfig.from_config(config_manager.get_cache_config())
    )

    def __init__(self, router: Processor, processor: Processor):
        self.router = router
        self.processor = processor

    def get_task_id(self, result: Dict[str, Any]) -> str:
        return result.get("data", {}).get("task_id", "")

    def is_async_task(self, result: Dict[str, Any]) -> str:
        """Determine if the result indicates an asynchronous task."""
        return (
            result.get("code") == 20000
            and result.get("status", False)
            and "task_id" in result.get("data", {})
        )

    def _get_result(self, result: Dict[str, Any], *, cache_key: str = None):
        try:
            response_data = result["data"]
            if BizyAirTask.check_inputs(result):
                self.cache_manager.set(cache_key, result)
                bz_task = BizyAirTask.from_data(result, check_inputs=False)
                bz_task.do_task_until_completed(timeout=10 * 60)  # 10 minutes
                last_data = bz_task.get_last_data()
                response_data = last_data.get("data")
            out = response_data["payload"]
            assert out is not None, "Output payload should not be None"
            self.cache_manager.set(cache_key, out, overwrite=True)
            return out
        except Exception as e:
            self.cache_manager.delete(cache_key)
            raise RuntimeError(f"Exception: {e}, response_data: {response_data}") from e

    def execute(
        self,
        prompt: Dict[str, Dict[str, Any]],
        last_node_ids: List[str],
        *args,
        **kwargs,
    ):

        prompt = encode_data(prompt)

        if BIZYAIR_DEBUG:
            debug_info = {
                "prompt": truncate_long_strings(prompt, 50),
                "last_node_ids": last_node_ids,
            }
            pprint.pprint(debug_info, indent=4)

        url = self.router(prompt=prompt, last_node_ids=last_node_ids)

        if BIZYAIR_DEBUG:
            print(f"Generated URL: {url}")

        start_time = time.time()
        sh256 = hashlib.sha256(
            json.dumps({"url": url, "prompt": prompt}).encode("utf-8")
        ).hexdigest()
        end_time = time.time()
        if BIZYAIR_DEBUG:
            print(
                f"Time taken to generate sh256-{sh256}: {end_time - start_time} seconds"
            )

        cached_output = self.cache_manager.get(sh256)
        if cached_output:
            if BIZYAIR_DEBUG:
                print(f"Cache hit for sh256-{sh256}")
            out = cached_output
        else:
            result = self.processor(url, prompt=prompt, last_node_ids=last_node_ids)
            out = self._get_result(result, cache_key=sh256)

        if BIZYAIR_DEBUG:
            pprint.pprint({"out": truncate_long_strings(out, 50)}, indent=4)

        try:
            real_out = decode_data(out)
            return [x[0] for x in real_out]
        except Exception as e:
            print("Exception occurred while decoding data")
            self.cache_manager.delete(sh256)
            traceback.print_exc()
            raise RuntimeError(f"Exception: {e=}") from e



import math
import os

import folder_paths
import torch
from bizyengine.core import BizyAirBaseNode, BizyAirNodeIO, create_node_data
from bizyengine.core.data_types import CLIP, CONDITIONING, MODEL

# set the models directory
if "ipadapter" not in folder_paths.folder_names_and_paths:
    current_paths = [os.path.join(folder_paths.models_dir, "ipadapter")]
else:
    current_paths, _ = folder_paths.folder_names_and_paths["ipadapter"]
folder_paths.folder_names_and_paths["ipadapter"] = (
    current_paths,
    folder_paths.supported_pt_extensions,
)

WEIGHT_TYPES = [
    "linear",
    "ease in",
    "ease out",
    "ease in-out",
    "reverse in-out",
    "weak input",
    "weak output",
    "weak middle",
    "strong middle",
    "style transfer",
    "composition",
    "strong style transfer",
    "style and composition",
    "style transfer precise",
    "composition precise",
]


# """
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Loaders
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# """


class IPAdapterUnifiedLoader(BizyAirBaseNode):
    def __init__(self):
        super().__init__()
        self.lora = None
        self.clipvision = {"file": None, "model": None}
        self.ipadapter = {"file": None, "model": None}
        self.insightface = {"provider": None, "model": None}

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (MODEL,),
                "preset": (
                    [
                        # "LIGHT - SD1.5 only (low strength)",
                        # "STANDARD (medium strength)",
                        # "VIT-G (medium strength)",
                        "PLUS (high strength)",
                        "PLUS FACE (portraits)",
                        # "FULL FACE - SD1.5 only (portraits stronger)",
                    ],
                ),
            },
            "optional": {
                "ipadapter": ("IPADAPTER",),
            },
        }

    NODE_DISPLAY_NAME = "IPAdapter Unified Loader"
    RETURN_TYPES = (
        MODEL,
        "IPADAPTER",
    )
    RETURN_NAMES = (
        "model",
        "ipadapter",
    )
    FUNCTION = "load_models"
    CATEGORY = "ipadapter"

    def load_models(self, **kwargs):
        assert kwargs.get("ipadapter", None) is None, "TODO"

        new_model: BizyAirNodeIO = kwargs["model"].copy(self.assigned_id)
        new_model.add_node_data(
            class_type="IPAdapterUnifiedLoader",
            inputs=kwargs,
            outputs={"slot_index": 0},
        )
        ipadapter: BizyAirNodeIO = kwargs["model"].copy(self.assigned_id)
        ipadapter.add_node_data(
            class_type="IPAdapterUnifiedLoader",
            inputs=kwargs,
            outputs={"slot_index": 1},
        )
        return (
            new_model,
            ipadapter,
        )


# class IPAdapterUnifiedLoaderFaceID(IPAdapterUnifiedLoader):
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "preset": (
#                     [
#                         "FACEID",
#                         "FACEID PLUS - SD1.5 only",
#                         "FACEID PLUS V2",
#                         "FACEID PORTRAIT (style transfer)",
#                         "FACEID PORTRAIT UNNORM - SDXL only (strong)",
#                     ],
#                 ),
#                 "lora_strength": (
#                     "FLOAT",
#                     {"default": 0.6, "min": 0, "max": 1, "step": 0.01},
#                 ),
#                 "provider": (
#                     ["CPU", "CUDA", "ROCM", "DirectML", "OpenVINO", "CoreML"],
#                 ),
#             },
#             "optional": {
#                 "ipadapter": ("IPADAPTER",),
#             },
#         }

#     RETURN_NAMES = (
#         "MODEL",
#         "ipadapter",
#     )
#     CATEGORY = "ipadapter/faceid"


# class IPAdapterUnifiedLoaderCommunity(IPAdapterUnifiedLoader):
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "preset": (["Composition", "Kolors"],),
#             },
#             "optional": {
#                 "ipadapter": ("IPADAPTER",),
#             },
#         }

#     CATEGORY = "ipadapter/loaders"


class IPAdapterModelLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {"ipadapter_file": (["kolors/ip_adapter_plus_general.bin"],)}
        }

    RETURN_TYPES = ("IPADAPTER",)
    FUNCTION = "load_ipadapter_model"
    CATEGORY = "ipadapter/loaders"

    def load_ipadapter_model(self, **kwargs):
        node_data = create_node_data(
            class_type="IPAdapterModelLoader",
            inputs=kwargs,
            outputs={"slot_index": 0},
        )
        return (BizyAirNodeIO(self.assigned_id, nodes={self.assigned_id: node_data}),)


# class IPAdapterInsightFaceLoader:
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "provider": (["CPU", "CUDA", "ROCM"],),
#             },
#         }

#     RETURN_TYPES = ("INSIGHTFACE",)
#     FUNCTION = "load_insightface"
#     CATEGORY = "ipadapter/loaders"


# """
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Main Apply Nodes
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# """


class IPAdapterSimple(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (MODEL,),
                "ipadapter": ("IPADAPTER",),
                "image": ("IMAGE",),
                "weight": (
                    "FLOAT",
                    {"default": 1.0, "min": -1, "max": 3, "step": 0.05},
                ),
                "start_at": (
                    "FLOAT",
                    {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
                ),
                "end_at": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
                ),
                "weight_type": (
                    ["standard", "prompt is more important", "style transfer"],
                ),
            },
            "optional": {
                "attn_mask": ("MASK",),
            },
        }

    NODE_DISPLAY_NAME = "IPAdapter"
    RETURN_TYPES = (MODEL,)
    RETURN_NAMES = "model"
    FUNCTION = "apply_ipadapter"
    CATEGORY = "ipadapter"

    def apply_ipadapter(self, **kwargs):
        new_model: BizyAirNodeIO = kwargs["model"].copy(self.assigned_id)
        new_model.add_node_data(
            class_type="IPAdapter",
            inputs=kwargs,
            outputs={"slot_index": 0},
        )
        return (new_model,)


class IPAdapterAdvanced(BizyAirBaseNode):
    def __init__(self):
        super().__init__()
        self.unfold_batch = False

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (MODEL,),
                "ipadapter": ("IPADAPTER",),
                "image": ("IMAGE",),
                "weight": (
                    "FLOAT",
                    {"default": 1.0, "min": -1, "max": 5, "step": 0.05},
                ),
                "weight_type": (WEIGHT_TYPES,),
                "combine_embeds": (
                    ["concat", "add", "subtract", "average", "norm average"],
                ),
                "start_at": (
                    "FLOAT",
                    {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
                ),
                "end_at": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
                ),
                "embeds_scaling": (
                    ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
                ),
            },
            "optional": {
                "image_negative": ("IMAGE",),
                "attn_mask": ("MASK",),
                "clip_vision": ("CLIP_VISION",),
            },
        }

    RETURN_TYPES = (MODEL,)
    RETURN_NAMES = ("model",)
    FUNCTION = "apply_ipadapter"
    CATEGORY = "ipadapter"

    def apply_ipadapter(self, **kwargs):
        new_model: BizyAirNodeIO = kwargs["model"].copy(self.assigned_id)
        new_model.add_node_data(
            class_type=self.__class__.__name__, inputs=kwargs, outputs={"slot_index": 0}
        )
        return (new_model,)


# class IPAdapterBatch(IPAdapterAdvanced):
#     def __init__(self):
#         super().__init__()
#         self.unfold_batch = True

#     NODE_DISPLAY_NAME = "IPAdapter Batch (Adv.)"

#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "ipadapter": ("IPADAPTER",),
#                 "image": ("IMAGE",),
#                 "weight": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 5, "step": 0.05},
#                 ),
#                 "weight_type": (WEIGHT_TYPES,),
#                 "start_at": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "end_at": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "embeds_scaling": (
#                     ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
#                 ),
#                 "encode_batch_size": ("INT", {"default": 0, "min": 0, "max": 4096}),
#             },
#             "optional": {
#                 "image_negative": ("IMAGE",),
#                 "attn_mask": ("MASK",),
#                 "clip_vision": ("CLIP_VISION",),
#             },
#         }


class IPAdapterStyleComposition(IPAdapterAdvanced):
    def __init__(self):
        super().__init__()

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (MODEL,),
                "ipadapter": ("IPADAPTER",),
                "image_style": ("IMAGE",),
                "image_composition": ("IMAGE",),
                "weight_style": (
                    "FLOAT",
                    {"default": 1.0, "min": -1, "max": 5, "step": 0.05},
                ),
                "weight_composition": (
                    "FLOAT",
                    {"default": 1.0, "min": -1, "max": 5, "step": 0.05},
                ),
                "expand_style": ("BOOLEAN", {"default": False}),
                "combine_embeds": (
                    ["concat", "add", "subtract", "average", "norm average"],
                    {"default": "average"},
                ),
                "start_at": (
                    "FLOAT",
                    {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
                ),
                "end_at": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
                ),
                "embeds_scaling": (
                    ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
                ),
            },
            "optional": {
                "image_negative": ("IMAGE",),
                "attn_mask": ("MASK",),
                "clip_vision": ("CLIP_VISION",),
            },
        }

    CATEGORY = "ipadapter/style_composition"


# class IPAdapterStyleCompositionBatch(IPAdapterStyleComposition):
#     def __init__(self):
#         super().__init__()
#         self.unfold_batch = True

#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "ipadapter": ("IPADAPTER",),
#                 "image_style": ("IMAGE",),
#                 "image_composition": ("IMAGE",),
#                 "weight_style": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 5, "step": 0.05},
#                 ),
#                 "weight_composition": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 5, "step": 0.05},
#                 ),
#                 "expand_style": ("BOOLEAN", {"default": False}),
#                 "start_at": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "end_at": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "embeds_scaling": (
#                     ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
#                 ),
#             },
#             "optional": {
#                 "image_negative": ("IMAGE",),
#                 "attn_mask": ("MASK",),
#                 "clip_vision": ("CLIP_VISION",),
#             },
#         }

#     NODE_DISPLAY_NAME = "IPAdapter Style & Composition Batch SDXL"


# class IPAdapterFaceID(IPAdapterAdvanced):
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "ipadapter": ("IPADAPTER",),
#                 "image": ("IMAGE",),
#                 "weight": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 3, "step": 0.05},
#                 ),
#                 "weight_faceidv2": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 5.0, "step": 0.05},
#                 ),
#                 "weight_type": (WEIGHT_TYPES,),
#                 "combine_embeds": (
#                     ["concat", "add", "subtract", "average", "norm average"],
#                 ),
#                 "start_at": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "end_at": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "embeds_scaling": (
#                     ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
#                 ),
#             },
#             "optional": {
#                 "image_negative": ("IMAGE",),
#                 "attn_mask": ("MASK",),
#                 "clip_vision": ("CLIP_VISION",),
#                 "insightface": ("INSIGHTFACE",),
#             },
#         }

#     CATEGORY = "ipadapter/faceid"
#     RETURN_TYPES = (
#         "MODEL",
#         "IMAGE",
#     )
#     RETURN_NAMES = (
#         "MODEL",
#         "face_image",
#     )
#     NODE_DISPLAY_NAME = "IPAdapter FaceID"


# class IPAAdapterFaceIDBatch(IPAdapterFaceID):
#     def __init__(self):
#         super().__init__()
#         self.unfold_batch = True

#     NODE_DISPLAY_NAME = "IPAdapter FaceID Batch"


# class IPAdapterTiled:
#     def __init__(self):
#         self.unfold_batch = False

#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "ipadapter": ("IPADAPTER",),
#                 "image": ("IMAGE",),
#                 "weight": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 3, "step": 0.05},
#                 ),
#                 "weight_type": (WEIGHT_TYPES,),
#                 "combine_embeds": (
#                     ["concat", "add", "subtract", "average", "norm average"],
#                 ),
#                 "start_at": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "end_at": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "sharpening": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.05},
#                 ),
#                 "embeds_scaling": (
#                     ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
#                 ),
#             },
#             "optional": {
#                 "image_negative": ("IMAGE",),
#                 "attn_mask": ("MASK",),
#                 "clip_vision": ("CLIP_VISION",),
#             },
#         }

#     RETURN_TYPES = (
#         "MODEL",
#         "IMAGE",
#         "MASK",
#     )
#     RETURN_NAMES = (
#         "MODEL",
#         "tiles",
#         "masks",
#     )
#     FUNCTION = "apply_tiled"
#     CATEGORY = "ipadapter/tiled"

#     def apply_tiled(
#         self,
#         model,
#         ipadapter,
#         image,
#         weight,
#         weight_type,
#         start_at,
#         end_at,
#         sharpening,
#         combine_embeds="concat",
#         image_negative=None,
#         attn_mask=None,
#         clip_vision=None,
#         embeds_scaling="V only",
#         encode_batch_size=0,
#     ):
#         # 1. Select the models
#         if "ipadapter" in ipadapter:
#             ipadapter_model = ipadapter["ipadapter"]["model"]
#             clip_vision = (
#                 clip_vision
#                 if clip_vision is not None
#                 else ipadapter["clipvision"]["model"]
#             )
#         else:
#             ipadapter_model = ipadapter
#             clip_vision = clip_vision

#         if clip_vision is None:
#             raise Exception("Missing CLIPVision model.")

#         del ipadapter

#         # 2. Extract the tiles
#         tile_size = 256  # I'm using 256 instead of 224 as it is more likely divisible by the latent size, it will be downscaled to 224 by the clip vision encoder
#         _, oh, ow, _ = image.shape
#         if attn_mask is None:
#             attn_mask = torch.ones([1, oh, ow], dtype=image.dtype, device=image.device)

#         image = image.permute([0, 3, 1, 2])
#         attn_mask = attn_mask.unsqueeze(1)
#         # the mask should have the same proportions as the reference image and the latent
#         attn_mask = T.Resize(
#             (oh, ow), interpolation=T.InterpolationMode.BICUBIC, antialias=True
#         )(attn_mask)

#         # if the image is almost a square, we crop it to a square
#         if oh / ow > 0.75 and oh / ow < 1.33:
#             # crop the image to a square
#             image = T.CenterCrop(min(oh, ow))(image)
#             resize = (tile_size * 2, tile_size * 2)

#             attn_mask = T.CenterCrop(min(oh, ow))(attn_mask)
#         # otherwise resize the smallest side and the other proportionally
#         else:
#             resize = (
#                 (int(tile_size * ow / oh), tile_size)
#                 if oh < ow
#                 else (tile_size, int(tile_size * oh / ow))
#             )

#         # using PIL for better results
#         imgs = []
#         for img in image:
#             img = T.ToPILImage()(img)
#             img = img.resize(resize, resample=Image.Resampling["LANCZOS"])
#             imgs.append(T.ToTensor()(img))
#         image = torch.stack(imgs)
#         del imgs, img

#         # we don't need a high quality resize for the mask
#         attn_mask = T.Resize(
#             resize[::-1], interpolation=T.InterpolationMode.BICUBIC, antialias=True
#         )(attn_mask)

#         # we allow a maximum of 4 tiles
#         if oh / ow > 4 or oh / ow < 0.25:
#             crop = (tile_size, tile_size * 4) if oh < ow else (tile_size * 4, tile_size)
#             image = T.CenterCrop(crop)(image)
#             attn_mask = T.CenterCrop(crop)(attn_mask)

#         attn_mask = attn_mask.squeeze(1)

#         if sharpening > 0:
#             image = contrast_adaptive_sharpening(image, sharpening)

#         image = image.permute([0, 2, 3, 1])

#         _, oh, ow, _ = image.shape

#         # find the number of tiles for each side
#         tiles_x = math.ceil(ow / tile_size)
#         tiles_y = math.ceil(oh / tile_size)
#         overlap_x = max(
#             0, (tiles_x * tile_size - ow) / (tiles_x - 1 if tiles_x > 1 else 1)
#         )
#         overlap_y = max(
#             0, (tiles_y * tile_size - oh) / (tiles_y - 1 if tiles_y > 1 else 1)
#         )

#         base_mask = torch.zeros(
#             [attn_mask.shape[0], oh, ow], dtype=image.dtype, device=image.device
#         )

#         # extract all the tiles from the image and create the masks
#         tiles = []
#         masks = []
#         for y in range(tiles_y):
#             for x in range(tiles_x):
#                 start_x = int(x * (tile_size - overlap_x))
#                 start_y = int(y * (tile_size - overlap_y))
#                 tiles.append(
#                     image[
#                         :,
#                         start_y : start_y + tile_size,
#                         start_x : start_x + tile_size,
#                         :,
#                     ]
#                 )
#                 mask = base_mask.clone()
#                 mask[
#                     :, start_y : start_y + tile_size, start_x : start_x + tile_size
#                 ] = attn_mask[
#                     :, start_y : start_y + tile_size, start_x : start_x + tile_size
#                 ]
#                 masks.append(mask)
#         del mask

#         # 3. Apply the ipadapter to each group of tiles
#         model = model.clone()
#         for i in range(len(tiles)):
#             ipa_args = {
#                 "image": tiles[i],
#                 "image_negative": image_negative,
#                 "weight": weight,
#                 "weight_type": weight_type,
#                 "combine_embeds": combine_embeds,
#                 "start_at": start_at,
#                 "end_at": end_at,
#                 "attn_mask": masks[i],
#                 "unfold_batch": self.unfold_batch,
#                 "embeds_scaling": embeds_scaling,
#                 "encode_batch_size": encode_batch_size,
#             }
#             # apply the ipadapter to the model without cloning it
#             model, _ = ipadapter_execute(
#                 model, ipadapter_model, clip_vision, **ipa_args
#             )

#         return (
#             model,
#             torch.cat(tiles),
#             torch.cat(masks),
#         )


# class IPAdapterTiledBatch(IPAdapterTiled):
#     def __init__(self):
#         self.unfold_batch = True

#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "ipadapter": ("IPADAPTER",),
#                 "image": ("IMAGE",),
#                 "weight": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 3, "step": 0.05},
#                 ),
#                 "weight_type": (WEIGHT_TYPES,),
#                 "start_at": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "end_at": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "sharpening": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.05},
#                 ),
#                 "embeds_scaling": (
#                     ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
#                 ),
#                 "encode_batch_size": ("INT", {"default": 0, "min": 0, "max": 4096}),
#             },
#             "optional": {
#                 "image_negative": ("IMAGE",),
#                 "attn_mask": ("MASK",),
#                 "clip_vision": ("CLIP_VISION",),
#             },
#         }


# class IPAdapterEmbeds:
#     def __init__(self):
#         self.unfold_batch = False

#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "ipadapter": ("IPADAPTER",),
#                 "pos_embed": ("EMBEDS",),
#                 "weight": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 3, "step": 0.05},
#                 ),
#                 "weight_type": (WEIGHT_TYPES,),
#                 "start_at": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "end_at": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "embeds_scaling": (
#                     ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
#                 ),
#             },
#             "optional": {
#                 "neg_embed": ("EMBEDS",),
#                 "attn_mask": ("MASK",),
#                 "clip_vision": ("CLIP_VISION",),
#             },
#         }

#     RETURN_TYPES = ("MODEL",)
#     FUNCTION = "apply_ipadapter"
#     CATEGORY = "ipadapter/embeds"

#     def apply_ipadapter(
#         self,
#         model,
#         ipadapter,
#         pos_embed,
#         weight,
#         weight_type,
#         start_at,
#         end_at,
#         neg_embed=None,
#         attn_mask=None,
#         clip_vision=None,
#         embeds_scaling="V only",
#     ):
#         ipa_args = {
#             "pos_embed": pos_embed,
#             "neg_embed": neg_embed,
#             "weight": weight,
#             "weight_type": weight_type,
#             "start_at": start_at,
#             "end_at": end_at,
#             "attn_mask": attn_mask,
#             "embeds_scaling": embeds_scaling,
#             "unfold_batch": self.unfold_batch,
#         }

#         if "ipadapter" in ipadapter:
#             ipadapter_model = ipadapter["ipadapter"]["model"]
#             clip_vision = (
#                 clip_vision
#                 if clip_vision is not None
#                 else ipadapter["clipvision"]["model"]
#             )
#         else:
#             ipadapter_model = ipadapter
#             clip_vision = clip_vision

#         if clip_vision is None and neg_embed is None:
#             raise Exception("Missing CLIPVision model.")

#         del ipadapter

#         return ipadapter_execute(
#             model.clone(), ipadapter_model, clip_vision, **ipa_args
#         )


# class IPAdapterEmbedsBatch(IPAdapterEmbeds):
#     def __init__(self):
#         self.unfold_batch = True


# class IPAdapterMS(IPAdapterAdvanced):
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "ipadapter": ("IPADAPTER",),
#                 "image": ("IMAGE",),
#                 "weight": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 5, "step": 0.05},
#                 ),
#                 "weight_faceidv2": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 5.0, "step": 0.05},
#                 ),
#                 "weight_type": (WEIGHT_TYPES,),
#                 "combine_embeds": (
#                     ["concat", "add", "subtract", "average", "norm average"],
#                 ),
#                 "start_at": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "end_at": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "embeds_scaling": (
#                     ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
#                 ),
#                 "layer_weights": ("STRING", {"default": "", "multiline": True}),
#             },
#             "optional": {
#                 "image_negative": ("IMAGE",),
#                 "attn_mask": ("MASK",),
#                 "clip_vision": ("CLIP_VISION",),
#                 "insightface": ("INSIGHTFACE",),
#             },
#         }

#     CATEGORY = "ipadapter/dev"


# class IPAdapterClipVisionEnhancer(IPAdapterAdvanced):
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "ipadapter": ("IPADAPTER",),
#                 "image": ("IMAGE",),
#                 "weight": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 5, "step": 0.05},
#                 ),
#                 "weight_type": (WEIGHT_TYPES,),
#                 "combine_embeds": (
#                     ["concat", "add", "subtract", "average", "norm average"],
#                 ),
#                 "start_at": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "end_at": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "embeds_scaling": (
#                     ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
#                 ),
#                 "enhance_tiles": ("INT", {"default": 2, "min": 1, "max": 16}),
#                 "enhance_ratio": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.05},
#                 ),
#             },
#             "optional": {
#                 "image_negative": ("IMAGE",),
#                 "attn_mask": ("MASK",),
#                 "clip_vision": ("CLIP_VISION",),
#             },
#         }

#     CATEGORY = "ipadapter/dev"


# class IPAdapterFromParams(IPAdapterAdvanced):
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "ipadapter": ("IPADAPTER",),
#                 "ipadapter_params": ("IPADAPTER_PARAMS",),
#                 "combine_embeds": (
#                     ["concat", "add", "subtract", "average", "norm average"],
#                 ),
#                 "embeds_scaling": (
#                     ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
#                 ),
#             },
#             "optional": {
#                 "image_negative": ("IMAGE",),
#                 "clip_vision": ("CLIP_VISION",),
#             },
#         }

#     CATEGORY = "ipadapter/params"


# class IPAdapterPreciseStyleTransfer(IPAdapterAdvanced):
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "ipadapter": ("IPADAPTER",),
#                 "image": ("IMAGE",),
#                 "weight": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 5, "step": 0.05},
#                 ),
#                 "style_boost": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -5, "max": 5, "step": 0.05},
#                 ),
#                 "combine_embeds": (
#                     ["concat", "add", "subtract", "average", "norm average"],
#                 ),
#                 "start_at": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "end_at": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "embeds_scaling": (
#                     ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
#                 ),
#             },
#             "optional": {
#                 "image_negative": ("IMAGE",),
#                 "attn_mask": ("MASK",),
#                 "clip_vision": ("CLIP_VISION",),
#             },
#         }


# class IPAdapterPreciseStyleTransferBatch(IPAdapterPreciseStyleTransfer):
#     def __init__(self):
#         self.unfold_batch = True


# class IPAdapterPreciseComposition(IPAdapterAdvanced):
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "model": ("MODEL",),
#                 "ipadapter": ("IPADAPTER",),
#                 "image": ("IMAGE",),
#                 "weight": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1, "max": 5, "step": 0.05},
#                 ),
#                 "composition_boost": (
#                     "FLOAT",
#                     {"default": 0.0, "min": -5, "max": 5, "step": 0.05},
#                 ),
#                 "combine_embeds": (
#                     ["concat", "add", "subtract", "average", "norm average"],
#                 ),
#                 "start_at": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "end_at": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "embeds_scaling": (
#                     ["V only", "K+V", "K+V w/ C penalty", "K+mean(V) w/ C penalty"],
#                 ),
#             },
#             "optional": {
#                 "image_negative": ("IMAGE",),
#                 "attn_mask": ("MASK",),
#                 "clip_vision": ("CLIP_VISION",),
#             },
#         }


# class IPAdapterPreciseCompositionBatch(IPAdapterPreciseComposition):
#     def __init__(self):
#         self.unfold_batch = True


# """
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Helpers
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# """


# class IPAdapterEncoder:
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "ipadapter": ("IPADAPTER",),
#                 "image": ("IMAGE",),
#                 "weight": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1.0, "max": 3.0, "step": 0.01},
#                 ),
#             },
#             "optional": {
#                 "mask": ("MASK",),
#                 "clip_vision": ("CLIP_VISION",),
#             },
#         }

#     RETURN_TYPES = (
#         "EMBEDS",
#         "EMBEDS",
#     )
#     RETURN_NAMES = (
#         "pos_embed",
#         "neg_embed",
#     )
#     FUNCTION = "encode"
#     CATEGORY = "ipadapter/embeds"


# class IPAdapterCombineEmbeds:
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "embed1": ("EMBEDS",),
#                 "method": (
#                     [
#                         "concat",
#                         "add",
#                         "subtract",
#                         "average",
#                         "norm average",
#                         "max",
#                         "min",
#                     ],
#                 ),
#             },
#             "optional": {
#                 "embed2": ("EMBEDS",),
#                 "embed3": ("EMBEDS",),
#                 "embed4": ("EMBEDS",),
#                 "embed5": ("EMBEDS",),
#             },
#         }

#     RETURN_TYPES = ("EMBEDS",)
#     FUNCTION = "batch"
#     CATEGORY = "ipadapter/embeds"

#     def batch(self, embed1, method, embed2=None, embed3=None, embed4=None, embed5=None):
#         if (
#             method == "concat"
#             and embed2 is None
#             and embed3 is None
#             and embed4 is None
#             and embed5 is None
#         ):
#             return (embed1,)

#         embeds = [embed1, embed2, embed3, embed4, embed5]
#         embeds = [embed for embed in embeds if embed is not None]
#         embeds = torch.cat(embeds, dim=0)

#         if method == "add":
#             embeds = torch.sum(embeds, dim=0).unsqueeze(0)
#         elif method == "subtract":
#             embeds = embeds[0] - torch.mean(embeds[1:], dim=0)
#             embeds = embeds.unsqueeze(0)
#         elif method == "average":
#             embeds = torch.mean(embeds, dim=0).unsqueeze(0)
#         elif method == "norm average":
#             embeds = torch.mean(
#                 embeds / torch.norm(embeds, dim=0, keepdim=True), dim=0
#             ).unsqueeze(0)
#         elif method == "max":
#             embeds = torch.max(embeds, dim=0).values.unsqueeze(0)
#         elif method == "min":
#             embeds = torch.min(embeds, dim=0).values.unsqueeze(0)

#         return (embeds,)


# class IPAdapterNoise:
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "type": (["fade", "dissolve", "gaussian", "shuffle"],),
#                 "strength": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0, "max": 1, "step": 0.05},
#                 ),
#                 "blur": ("INT", {"default": 0, "min": 0, "max": 32, "step": 1}),
#             },
#             "optional": {
#                 "image_optional": ("IMAGE",),
#             },
#         }

#     RETURN_TYPES = ("IMAGE",)
#     FUNCTION = "make_noise"
#     CATEGORY = "ipadapter/utils"

#     def make_noise(self, type, strength, blur, image_optional=None):
#         if image_optional is None:
#             image = torch.zeros([1, 224, 224, 3])
#         else:
#             transforms = T.Compose(
#                 [
#                     T.CenterCrop(min(image_optional.shape[1], image_optional.shape[2])),
#                     T.Resize(
#                         (224, 224),
#                         interpolation=T.InterpolationMode.BICUBIC,
#                         antialias=True,
#                     ),
#                 ]
#             )
#             image = transforms(image_optional.permute([0, 3, 1, 2])).permute(
#                 [0, 2, 3, 1]
#             )

#         seed = (
#             int(torch.sum(image).item()) % 1000000007
#         )  # hash the image to get a seed, grants predictability
#         torch.manual_seed(seed)

#         if type == "fade":
#             noise = torch.rand_like(image)
#             noise = image * (1 - strength) + noise * strength
#         elif type == "dissolve":
#             mask = (torch.rand_like(image) < strength).float()
#             noise = torch.rand_like(image)
#             noise = image * (1 - mask) + noise * mask
#         elif type == "gaussian":
#             noise = torch.randn_like(image) * strength
#             noise = image + noise
#         elif type == "shuffle":
#             transforms = T.Compose(
#                 [
#                     T.ElasticTransform(alpha=75.0, sigma=(1 - strength) * 3.5),
#                     T.RandomVerticalFlip(p=1.0),
#                     T.RandomHorizontalFlip(p=1.0),
#                 ]
#             )
#             image = transforms(image.permute([0, 3, 1, 2])).permute([0, 2, 3, 1])
#             noise = torch.randn_like(image) * (strength * 0.75)
#             noise = image * (1 - noise) + noise

#         del image
#         noise = torch.clamp(noise, 0, 1)

#         if blur > 0:
#             if blur % 2 == 0:
#                 blur += 1
#             noise = T.functional.gaussian_blur(
#                 noise.permute([0, 3, 1, 2]), blur
#             ).permute([0, 2, 3, 1])

#         return (noise,)


# class PrepImageForClipVision:
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "image": ("IMAGE",),
#                 "interpolation": (
#                     ["LANCZOS", "BICUBIC", "HAMMING", "BILINEAR", "BOX", "NEAREST"],
#                 ),
#                 "crop_position": (["top", "bottom", "left", "right", "center", "pad"],),
#                 "sharpening": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0, "max": 1, "step": 0.05},
#                 ),
#             },
#         }

#     RETURN_TYPES = ("IMAGE",)
#     FUNCTION = "prep_image"

#     CATEGORY = "ipadapter/utils"

#     def prep_image(
#         self, image, interpolation="LANCZOS", crop_position="center", sharpening=0.0
#     ):
#         size = (224, 224)
#         _, oh, ow, _ = image.shape
#         output = image.permute([0, 3, 1, 2])

#         if crop_position == "pad":
#             if oh != ow:
#                 if oh > ow:
#                     pad = (oh - ow) // 2
#                     pad = (pad, 0, pad, 0)
#                 elif ow > oh:
#                     pad = (ow - oh) // 2
#                     pad = (0, pad, 0, pad)
#                 output = T.functional.pad(output, pad, fill=0)
#         else:
#             crop_size = min(oh, ow)
#             x = (ow - crop_size) // 2
#             y = (oh - crop_size) // 2
#             if "top" in crop_position:
#                 y = 0
#             elif "bottom" in crop_position:
#                 y = oh - crop_size
#             elif "left" in crop_position:
#                 x = 0
#             elif "right" in crop_position:
#                 x = ow - crop_size

#             x2 = x + crop_size
#             y2 = y + crop_size

#             output = output[:, :, y:y2, x:x2]

#         imgs = []
#         for img in output:
#             img = T.ToPILImage()(img)  # using PIL for better results
#             img = img.resize(size, resample=Image.Resampling[interpolation])
#             imgs.append(T.ToTensor()(img))
#         output = torch.stack(imgs, dim=0)
#         del imgs, img

#         if sharpening > 0:
#             output = contrast_adaptive_sharpening(output, sharpening)

#         output = output.permute([0, 2, 3, 1])

#         return (output,)


# class IPAdapterSaveEmbeds:
#     def __init__(self):
#         self.output_dir = folder_paths.get_output_directory()

#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "embeds": ("EMBEDS",),
#                 "filename_prefix": ("STRING", {"default": "IP_embeds"}),
#             },
#         }

#     RETURN_TYPES = ()
#     FUNCTION = "save"
#     OUTPUT_NODE = True
#     CATEGORY = "ipadapter/embeds"

#     def save(self, embeds, filename_prefix):
#         (
#             full_output_folder,
#             filename,
#             counter,
#             subfolder,
#             filename_prefix,
#         ) = folder_paths.get_save_image_path(filename_prefix, self.output_dir)
#         file = f"{filename}_{counter:05}.ipadpt"
#         file = os.path.join(full_output_folder, file)

#         torch.save(embeds, file)
#         return (None,)


# class IPAdapterLoadEmbeds:
#     @classmethod
#     def INPUT_TYPES(s):
#         input_dir = folder_paths.get_input_directory()
#         files = [
#             os.path.relpath(os.path.join(root, file), input_dir)
#             for root, dirs, files in os.walk(input_dir)
#             for file in files
#             if file.endswith(".ipadpt")
#         ]
#         return {
#             "required": {
#                 "embeds": [
#                     sorted(files),
#                 ]
#             },
#         }

#     RETURN_TYPES = ("EMBEDS",)
#     FUNCTION = "load"
#     CATEGORY = "ipadapter/embeds"

#     def load(self, embeds):
#         path = folder_paths.get_annotated_filepath(embeds)
#         return (torch.load(path).cpu(),)


# class IPAdapterWeights:
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "weights": ("STRING", {"default": "1.0, 0.0", "multiline": True}),
#                 "timing": (
#                     [
#                         "custom",
#                         "linear",
#                         "ease_in_out",
#                         "ease_in",
#                         "ease_out",
#                         "random",
#                     ],
#                     {"default": "linear"},
#                 ),
#                 "frames": ("INT", {"default": 0, "min": 0, "max": 9999, "step": 1}),
#                 "start_frame": (
#                     "INT",
#                     {"default": 0, "min": 0, "max": 9999, "step": 1},
#                 ),
#                 "end_frame": (
#                     "INT",
#                     {"default": 9999, "min": 0, "max": 9999, "step": 1},
#                 ),
#                 "add_starting_frames": (
#                     "INT",
#                     {"default": 0, "min": 0, "max": 9999, "step": 1},
#                 ),
#                 "add_ending_frames": (
#                     "INT",
#                     {"default": 0, "min": 0, "max": 9999, "step": 1},
#                 ),
#                 "method": (
#                     ["full batch", "shift batches", "alternate batches"],
#                     {"default": "full batch"},
#                 ),
#             },
#             "optional": {
#                 "image": ("IMAGE",),
#             },
#         }

#     RETURN_TYPES = ("FLOAT", "FLOAT", "INT", "IMAGE", "IMAGE", "WEIGHTS_STRATEGY")
#     RETURN_NAMES = (
#         "weights",
#         "weights_invert",
#         "total_frames",
#         "image_1",
#         "image_2",
#         "weights_strategy",
#     )
#     FUNCTION = "weights"
#     CATEGORY = "ipadapter/weights"


# class IPAdapterWeightsFromStrategy(IPAdapterWeights):
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "weights_strategy": ("WEIGHTS_STRATEGY",),
#             },
#             "optional": {
#                 "image": ("IMAGE",),
#             },
#         }


# class IPAdapterPromptScheduleFromWeightsStrategy:
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "weights_strategy": ("WEIGHTS_STRATEGY",),
#                 "prompt": ("STRING", {"default": "", "multiline": True}),
#             }
#         }

#     RETURN_TYPES = ("STRING",)
#     RETURN_NAMES = ("prompt_schedule",)
#     FUNCTION = "prompt_schedule"
#     CATEGORY = "ipadapter/weights"


# class IPAdapterCombineWeights:
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "weights_1": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.05},
#                 ),
#                 "weights_2": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.05},
#                 ),
#             }
#         }

#     RETURN_TYPES = ("FLOAT", "INT")
#     RETURN_NAMES = ("weights", "count")
#     FUNCTION = "combine"
#     CATEGORY = "ipadapter/utils"


# class IPAdapterRegionalConditioning:
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 # "set_cond_area": (["default", "mask bounds"],),
#                 "image": ("IMAGE",),
#                 "image_weight": (
#                     "FLOAT",
#                     {"default": 1.0, "min": -1.0, "max": 3.0, "step": 0.05},
#                 ),
#                 "prompt_weight": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 10.0, "step": 0.05},
#                 ),
#                 "weight_type": (WEIGHT_TYPES,),
#                 "start_at": (
#                     "FLOAT",
#                     {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#                 "end_at": (
#                     "FLOAT",
#                     {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
#                 ),
#             },
#             "optional": {
#                 "mask": ("MASK",),
#                 "positive": ("CONDITIONING",),
#                 "negative": ("CONDITIONING",),
#             },
#         }

#     RETURN_TYPES = (
#         "IPADAPTER_PARAMS",
#         "CONDITIONING",
#         "CONDITIONING",
#     )
#     RETURN_NAMES = ("IPADAPTER_PARAMS", "POSITIVE", "NEGATIVE")
#     FUNCTION = "conditioning"

#     CATEGORY = "ipadapter/params"


# class IPAdapterCombineParams:
#     @classmethod
#     def INPUT_TYPES(s):
#         return {
#             "required": {
#                 "params_1": ("IPADAPTER_PARAMS",),
#                 "params_2": ("IPADAPTER_PARAMS",),
#             },
#             "optional": {
#                 "params_3": ("IPADAPTER_PARAMS",),
#                 "params_4": ("IPADAPTER_PARAMS",),
#                 "params_5": ("IPADAPTER_PARAMS",),
#             },
#         }

#     RETURN_TYPES = ("IPADAPTER_PARAMS",)
#     FUNCTION = "combine"
#     CATEGORY = "ipadapter/params"


# """
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Register
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# """

# NODE_CLASS_MAPPINGS = {
#     # Main Apply Nodes
#     "IPAdapter": IPAdapterSimple,
#     "IPAdapterAdvanced": IPAdapterAdvanced,
#     "IPAdapterBatch": IPAdapterBatch,
#     "IPAdapterFaceID": IPAdapterFaceID,
#     "IPAAdapterFaceIDBatch": IPAAdapterFaceIDBatch,
#     "IPAdapterTiled": IPAdapterTiled,
#     "IPAdapterTiledBatch": IPAdapterTiledBatch,
#     "IPAdapterEmbeds": IPAdapterEmbeds,
#     "IPAdapterEmbedsBatch": IPAdapterEmbedsBatch,
#     "IPAdapterStyleComposition": IPAdapterStyleComposition,
#     "IPAdapterStyleCompositionBatch": IPAdapterStyleCompositionBatch,
#     "IPAdapterMS": IPAdapterMS,
#     "IPAdapterClipVisionEnhancer": IPAdapterClipVisionEnhancer,
#     "IPAdapterFromParams": IPAdapterFromParams,
#     "IPAdapterPreciseStyleTransfer": IPAdapterPreciseStyleTransfer,
#     "IPAdapterPreciseStyleTransferBatch": IPAdapterPreciseStyleTransferBatch,
#     "IPAdapterPreciseComposition": IPAdapterPreciseComposition,
#     "IPAdapterPreciseCompositionBatch": IPAdapterPreciseCompositionBatch,
#     # Loaders
#     "IPAdapterUnifiedLoader": IPAdapterUnifiedLoader,
#     "IPAdapterUnifiedLoaderFaceID": IPAdapterUnifiedLoaderFaceID,
#     "IPAdapterModelLoader": IPAdapterModelLoader,
#     "IPAdapterInsightFaceLoader": IPAdapterInsightFaceLoader,
#     "IPAdapterUnifiedLoaderCommunity": IPAdapterUnifiedLoaderCommunity,
#     # Helpers
#     "IPAdapterEncoder": IPAdapterEncoder,
#     "IPAdapterCombineEmbeds": IPAdapterCombineEmbeds,
#     "IPAdapterNoise": IPAdapterNoise,
#     "PrepImageForClipVision": PrepImageForClipVision,
#     "IPAdapterSaveEmbeds": IPAdapterSaveEmbeds,
#     "IPAdapterLoadEmbeds": IPAdapterLoadEmbeds,
#     "IPAdapterWeights": IPAdapterWeights,
#     "IPAdapterCombineWeights": IPAdapterCombineWeights,
#     "IPAdapterWeightsFromStrategy": IPAdapterWeightsFromStrategy,
#     "IPAdapterPromptScheduleFromWeightsStrategy": IPAdapterPromptScheduleFromWeightsStrategy,
#     "IPAdapterRegionalConditioning": IPAdapterRegionalConditioning,
#     "IPAdapterCombineParams": IPAdapterCombineParams,
# }

# NODE_DISPLAY_NAME_MAPPINGS = {
#     # Main Apply Nodes
#     "IPAdapter": "IPAdapter",
#     "IPAdapterAdvanced": "IPAdapter Advanced",
#     "IPAdapterBatch": "IPAdapter Batch (Adv.)",
#     "IPAdapterFaceID": "IPAdapter FaceID",
#     "IPAAdapterFaceIDBatch": "IPAdapter FaceID Batch",
#     "IPAdapterTiled": "IPAdapter Tiled",
#     "IPAdapterTiledBatch": "IPAdapter Tiled Batch",
#     "IPAdapterEmbeds": "IPAdapter Embeds",
#     "IPAdapterEmbedsBatch": "IPAdapter Embeds Batch",
#     "IPAdapterStyleComposition": "IPAdapter Style & Composition SDXL",
#     "IPAdapterStyleCompositionBatch": "IPAdapter Style & Composition Batch SDXL",
#     "IPAdapterMS": "IPAdapter Mad Scientist",
#     "IPAdapterClipVisionEnhancer": "IPAdapter ClipVision Enhancer",
#     "IPAdapterFromParams": "IPAdapter from Params",
#     "IPAdapterPreciseStyleTransfer": "IPAdapter Precise Style Transfer",
#     "IPAdapterPreciseStyleTransferBatch": "IPAdapter Precise Style Transfer Batch",
#     "IPAdapterPreciseComposition": "IPAdapter Precise Composition",
#     "IPAdapterPreciseCompositionBatch": "IPAdapter Precise Composition Batch",
#     # Loaders
#     "IPAdapterUnifiedLoader": "IPAdapter Unified Loader",
#     "IPAdapterUnifiedLoaderFaceID": "IPAdapter Unified Loader FaceID",
#     "IPAdapterModelLoader": "IPAdapter Model Loader",
#     "IPAdapterInsightFaceLoader": "IPAdapter InsightFace Loader",
#     "IPAdapterUnifiedLoaderCommunity": "IPAdapter Unified Loader Community",
#     # Helpers
#     "IPAdapterEncoder": "IPAdapter Encoder",
#     "IPAdapterCombineEmbeds": "IPAdapter Combine Embeds",
#     "IPAdapterNoise": "IPAdapter Noise",
#     "PrepImageForClipVision": "Prep Image For ClipVision",
#     "IPAdapterSaveEmbeds": "IPAdapter Save Embeds",
#     "IPAdapterLoadEmbeds": "IPAdapter Load Embeds",
#     "IPAdapterWeights": "IPAdapter Weights",
#     "IPAdapterWeightsFromStrategy": "IPAdapter Weights From Strategy",
#     "IPAdapterPromptScheduleFromWeightsStrategy": "Prompt Schedule From Weights Strategy",
#     "IPAdapterCombineWeights": "IPAdapter Combine Weights",
#     "IPAdapterRegionalConditioning": "IPAdapter Regional Conditioning",
#     "IPAdapterCombineParams": "IPAdapter Combine Params",
# }



class ErrorNo:
    def __init__(self, http_status_code, code, payload, message):
        self.http_status_code = http_status_code
        self.code = code
        self.message = message
        self.data = payload

    def copy(self):
        return ErrorNo(self.http_status_code, self.code, self.data, self.message)


class errnos:
    OK = ErrorNo(200, 20000, None, "Success")
    NO_MODEL_FOUND = ErrorNo(404, 20226, None, "No model found")

    INVALID_TYPE = ErrorNo(400, 400100, None, "Invalid model type")
    INVALID_NAME = ErrorNo(400, 400101, None, "Invalid model name")
    NO_FILE_UPLOAD = ErrorNo(400, 400102, None, "No file to upload")
    EMPTY_UPLOAD_ID = ErrorNo(400, 400103, None, "Upload id is empty")
    INVALID_SHARE_ID = ErrorNo(400, 400104, None, "Invalid share id")
    EMPTY_ABS_PATH = ErrorNo(400, 400111, None, "The upload path cannot be empty")
    NO_ABS_PATH = ErrorNo(400, 400112, None, "The upload path is not an absolute path")
    PATH_NOT_EXISTS = ErrorNo(400, 400113, None, "The upload path does not exist")
    INVALID_CLIENT_ID = ErrorNo(400, 400114, None, "Invalid client id")
    NO_PUBLIC_FLAG = ErrorNo(
        400, 400115, None, 'The parameter "public" is not provided'
    )
    FILE_NOT_EXISTS = ErrorNo(400, 400116, None, "The file does not exist")
    NO_SHARE_ID = ErrorNo(400, 400117, None, 'The parameter "share_id" is not provided')
    INVALID_DESCRIPTION = ErrorNo(400, 400118, None, "Invalid description")
    NOT_A_FILE = ErrorNo(400, 400119, None, "The path is not a file")
    NOT_ALLOWED_EXT_NAME = ErrorNo(400, 400120, None, "Not allowed extension name")
    INVALID_UPLOAD_ID = ErrorNo(400, 400121, None, "Invalid upload id")
    INVALID_VERSIONS = ErrorNo(400, 400122, None, "Invalid versions")
    DUPLICATE_VERSION = ErrorNo(400, 400123, None, "Duplicate version")
    INVALID_VERSION_FIELD = ErrorNo(400, 400124, None, "Invalid version field")
    INVALID_QUERY_MODE = ErrorNo(400, 400125, None, "Invalid query mode")
    INVALID_VERSION_NAME = ErrorNo(400, 400126, None, "Invalid model version name")
    INVALID_MODEL_ID = ErrorNo(400, 400127, None, "Invalid model id")
    INVALID_MODEL_VERSION_ID = ErrorNo(400, 400128, None, "Invalid model version id")
    UNSUPPORT_LIKE_TYPE = ErrorNo(400, 400130, None, "Unsupport like type")
    INVALID_SIGN = ErrorNo(400, 400131, None, "Invalid file sign")
    EMPTY_SHA256SUM = ErrorNo(400, 400132, None, "Empty sha256sum")
    INVALID_OBJECT_KEY = ErrorNo(400, 400133, None, "Invalid object key")
    FAILED_TO_FETCH_WORKFLOW_JSON = ErrorNo(
        400, 400134, None, "Failed to fetch workflow json"
    )
    INVALID_DATASET_NAME = ErrorNo(400, 400135, None, "Invalid dataset name")
    INVALID_DATASET_VERSION = ErrorNo(400, 400136, None, "Invalid dataset version")
    INVALID_DATASET_ID = ErrorNo(400, 400137, None, "Invalid dataset id")
    INVALID_DATASET_VERSION_ID = ErrorNo(
        400, 400138, None, "Invalid dataset version id"
    )
    INVALID_SHARE_BIZ_ID = ErrorNo(400, 400139, None, "Invalid share biz id")
    INVALID_SHARE_TYPE = ErrorNo(400, 400140, None, "Invalid share type")
    INVALID_SHARE_CODE = ErrorNo(400, 400141, None, "Invalid share code")
    INVALID_NOTIF_ID = ErrorNo(400, 400142, None, "Invalid notification id")

    INVALID_API_KEY = ErrorNo(401, 401000, None, "Invalid API key")
    INVALID_USER = ErrorNo(401, 401001, None, "Invalid user")

    SIGN_FILE = ErrorNo(500, 500101, None, "Failed to sign file")
    UPLOAD = ErrorNo(500, 500102, None, "Failed to upload file")
    COMMIT_FILE = ErrorNo(500, 500103, None, "Failed to commit file")
    COMMIT_BIZY_MODEL = ErrorNo(500, 500105, None, "Failed to commit model")
    EMPTY_FILES = ErrorNo(500, 500107, None, "Empty files to make a model")
    LIST_MODEL_FILE = ErrorNo(500, 500108, None, "Failed to list model file")
    LIST_SHARE_MODEL_FILE = ErrorNo(
        500, 500108, None, "Failed to list share model file"
    )
    INVALID_FILENAME = ErrorNo(500, 500109, None, "Invalid filename")
    DELETE_MODEL = ErrorNo(500, 500110, None, "Failed to delete model")
    GET_USER_INFO = ErrorNo(500, 500111, None, "Failed to get user info")
    LIST_MODEL = ErrorNo(500, 500112, None, "Failed to list model")
    CHANGE_PUBLIC = ErrorNo(500, 500113, None, "Failed to change public")
    UPDATE_SHARE_ID = ErrorNo(500, 500114, None, "Failed to update share id")
    GET_DESCRIPTION = ErrorNo(500, 500115, None, "Failed to get description")
    UPDATE_DESCRIPTION = ErrorNo(500, 500116, None, "Failed to update description")
    DELETE_BIZY_MODEL = ErrorNo(500, 500117, None, "Failed to delete model")
    QUERY_COMMUNITY_MODELS = ErrorNo(
        500, 500118, None, "Failed to query community models"
    )
    QUERY_MODELS = ErrorNo(500, 500119, None, "Failed to query models")
    GET_MODEL_DETAIL = ErrorNo(500, 500120, None, "Failed to get model detail")
    GET_MODEL_VERSION_DETAIL = ErrorNo(
        500, 500121, None, "Failed to get model version detail"
    )
    FORK_MODEL_VERSION = ErrorNo(500, 500122, None, "Failed to fork model version")
    UPDATE_MODEL = ErrorNo(500, 500123, None, "Failed to update model")
    GET_UPLOAD_TOKEN = ErrorNo(500, 500124, None, "Failed to get upload token")
    TOGGLE_USER_LIKE = ErrorNo(500, 500125, None, "Failed to toggle user like")
    GET_DOWNLOAD_URL = ErrorNo(500, 500126, None, "Failed to get download url")
    DOWNLOAD_JSON = ErrorNo(500, 500127, None, "Failed to download json")
    LIST_SHARE_MODEL_FILE_ERR = ErrorNo(
        500, 500128, None, "Failed to list share model file"
    )
    QUERY_OFFICIAL_MODELS = ErrorNo(
        500, 500129, None, "Failed to query official models"
    )
    UNFORK_MODEL_VERSION = ErrorNo(500, 500130, None, "Failed to unfork model version")

    COMMIT_DATASET = ErrorNo(500, 500130, None, "Failed to commit dataset")
    UPDATE_DATASET = ErrorNo(500, 500131, None, "Failed to update dataset")
    GET_DATASET_VERSION_DETAIL = ErrorNo(
        500, 500132, None, "Failed to get dataset version detail"
    )
    DELETE_DATASET = ErrorNo(500, 500133, None, "Failed to delete dataset")
    QUERY_DATASETS = ErrorNo(500, 500134, None, "Failed to query datasets")
    GET_DATASET_DETAIL = ErrorNo(500, 500135, None, "Failed to get dataset detail")
    CREATE_SHARE = ErrorNo(500, 500136, None, "Failed to create share")
    GET_SHARE_DETAIL = ErrorNo(500, 500137, None, "Failed to get share detail")

    GET_DATA_DICT = ErrorNo(500, 500138, None, "Failed to get data dict")

    GET_NOTIF_UNREAD_COUNT = ErrorNo(
        500, 500139, None, "Failed to get notification unread counts"
    )
    QUERY_NOTIF = ErrorNo(500, 500140, None, "Failed to query notifications")
    READ_NOTIF = ErrorNo(500, 500141, None, "Failed to mark notifications as read")
    READ_ALL_NOTIF = ErrorNo(
        500, 500142, None, "Failed to mark all notifications as read"
    )



from bizyengine.core import BizyAirBaseNode

from .nodes_segment_anything_utils import *


class BizyAir_SAMModelLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "model_name": (list_sam_model(),),
            }
        }

    CATEGORY = "☁️BizyAir/segment-anything"
    # FUNCTION = "main"
    RETURN_TYPES = ("SAM_PREDICTOR",)
    NODE_DISPLAY_NAME = "☁️BizyAir Load SAM Model"


class BizyAir_GroundingDinoModelLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "model_name": (list_groundingdino_model(),),
            }
        }

    CATEGORY = "☁️BizyAir/segment-anything"
    # FUNCTION = "main"
    RETURN_TYPES = ("GROUNDING_DINO_MODEL",)
    NODE_DISPLAY_NAME = "☁️BizyAir Load GroundingDino Model"


class BizyAir_VITMatteModelLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(cls):
        method_list = [
            "VITMatte",
            "VITMatte(local)",
        ]
        return {
            "required": {
                "detail_method": (method_list,),
            }
        }

    CATEGORY = "☁️BizyAir/segment-anything"
    # FUNCTION = "main"
    RETURN_TYPES = (
        "VitMatte_MODEL",
        "VitMatte_predictor",
    )
    NODE_DISPLAY_NAME = "☁️BizyAir Load VITMatte Model"


class BizyAir_GroundingDinoSAMSegment(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "grounding_dino_model": ("GROUNDING_DINO_MODEL", {}),
                "sam_predictor": ("SAM_PREDICTOR", {}),
                "image": ("IMAGE", {}),
                "prompt": ("STRING", {}),
                "box_threshold": (
                    "FLOAT",
                    {"default": 0.3, "min": 0, "max": 1.0, "step": 0.01},
                ),
                "text_threshold": (
                    "FLOAT",
                    {"default": 0.3, "min": 0, "max": 1.0, "step": 0.01},
                ),
            }
        }

    CATEGORY = "☁️BizyAir/segment-anything"
    # FUNCTION = "main"
    RETURN_TYPES = ("IMAGE", "MASK")
    NODE_DISPLAY_NAME = "☁️BizyAir GroundingDinoSAMSegment"


class BizyAir_TrimapGenerate(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "mask": ("MASK",),
                "detail_erode": (
                    "INT",
                    {"default": 6, "min": 1, "max": 255, "step": 1},
                ),
                "detail_dilate": (
                    "INT",
                    {"default": 6, "min": 1, "max": 255, "step": 1},
                ),
            }
        }

    CATEGORY = "☁️BizyAir/segment-anything"
    # FUNCTION = "main"
    RETURN_TYPES = ("MASK",)
    RETURN_NAMES = ("trimap",)
    NODE_DISPLAY_NAME = "☁️BizyAir Trimap Generate"


class BizyAir_VITMattePredict(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "image": ("IMAGE", {}),
                "trimap": ("MASK",),
                "vitmatte_model": ("VitMatte_MODEL", {}),
                "vitmatte_predictor": ("VitMatte_predictor", {}),
                "black_point": (
                    "FLOAT",
                    {
                        "default": 0.15,
                        "min": 0.01,
                        "max": 0.98,
                        "step": 0.01,
                        "display": "slider",
                    },
                ),
                "white_point": (
                    "FLOAT",
                    {
                        "default": 0.99,
                        "min": 0.02,
                        "max": 0.99,
                        "step": 0.01,
                        "display": "slider",
                    },
                ),
                "max_megapixels": (
                    "FLOAT",
                    {"default": 2.0, "min": 1, "max": 999, "step": 0.1},
                ),
            }
        }

    CATEGORY = "☁️BizyAir/segment-anything"
    # FUNCTION = "main"
    RETURN_TYPES = (
        "IMAGE",
        "MASK",
    )
    RETURN_NAMES = (
        "image",
        "mask",
    )
    NODE_DISPLAY_NAME = "☁️BizyAir VITMatte Predict"


class BizyAirDetailMethodPredict(BizyAirBaseNode):
    NODE_DISPLAY_NAME = "☁️BizyAir DetailMethod Predict"

    @classmethod
    def INPUT_TYPES(cls):

        method_list = [
            "PyMatting",
        ]
        return {
            "required": {
                "image": ("IMAGE", {}),
                "mask": ("MASK",),
                "detail_method": (method_list,),
                "detail_erode": (
                    "INT",
                    {"default": 6, "min": 1, "max": 255, "step": 1},
                ),
                "detail_dilate": (
                    "INT",
                    {"default": 6, "min": 1, "max": 255, "step": 1},
                ),
                "black_point": (
                    "FLOAT",
                    {
                        "default": 0.15,
                        "min": 0.01,
                        "max": 0.98,
                        "step": 0.01,
                        "display": "slider",
                    },
                ),
                "white_point": (
                    "FLOAT",
                    {
                        "default": 0.99,
                        "min": 0.02,
                        "max": 0.99,
                        "step": 0.01,
                        "display": "slider",
                    },
                ),
            }
        }

    CATEGORY = "☁️BizyAir/segment-anything"
    FUNCTION = "main"
    RETURN_TYPES = (
        "IMAGE",
        "MASK",
    )
    RETURN_NAMES = (
        "image",
        "mask",
    )

    def main(
        self,
        image,
        mask,
        detail_method,
        detail_erode,
        detail_dilate,
        black_point,
        white_point,
    ):

        ret_images = []
        ret_masks = []
        # device = comfy.model_management.get_torch_device()

        for i in range(image.shape[0]):
            img = torch.unsqueeze(image[i], 0)
            img = pil2tensor(tensor2pil(img).convert("RGB"))
            _image = tensor2pil(img).convert("RGBA")

            detail_range = detail_erode + detail_dilate

            if detail_method == "PyMatting":
                _mask = tensor2pil(
                    mask_edge_detail(
                        img, mask[i], detail_range // 8 + 1, black_point, white_point
                    )
                )

            _image = RGB2RGBA(tensor2pil(img).convert("RGB"), _mask.convert("L"))

            ret_images.append(pil2tensor(_image))
            ret_masks.append(image2mask(_mask))
        if len(ret_masks) == 0:
            _, height, width, _ = image.size()
            empty_mask = torch.zeros(
                (1, height, width), dtype=torch.uint8, device="cpu"
            )
            return (empty_mask, empty_mask)

        return (
            torch.cat(ret_images, dim=0),
            torch.cat(ret_masks, dim=0),
        )



import json
import os
from enum import Enum

from aiohttp import web

BIZYAIR_DEBUG = os.getenv("BIZYAIR_DEBUG", False)
CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

SHOW_CASES = {}
SAM_COORDINATE = {}
IS_RESET_SAM = False


class EDIT_MODE(Enum):
    box = 0
    point = 1


from server import PromptServer


@PromptServer.instance.routes.post("/bizyair/postsam")
async def save_sam(request):
    global IS_RESET_SAM
    IS_RESET_SAM = False
    post = await request.post()
    SAM_COORDINATE["nums"] = post.get("nums")
    SAM_COORDINATE["mode"] = json.loads(post.get("mode"))
    if SAM_COORDINATE["mode"] == EDIT_MODE.point.value:
        SAM_COORDINATE["point_coords"] = json.loads(post.get("coords"))
    elif SAM_COORDINATE["mode"] == EDIT_MODE.box.value:
        SAM_COORDINATE["box_coords"] = json.loads(post.get("coords"))
    SAM_COORDINATE["filename"] = post.get("filename")

    return web.Response(status=200)


@PromptServer.instance.routes.get("/bizyair/getsam")
async def get_sam(request):
    return web.Response(
        text=json.dumps(SAM_COORDINATE, ensure_ascii=False),
        content_type="application/json",
    )


@PromptServer.instance.routes.get("/bizyair/resetsam")
async def reset_sam(request):
    global IS_RESET_SAM
    IS_RESET_SAM = True
    return web.Response(status=200)


@PromptServer.instance.routes.get("/bizyair/isresetsam")
async def isreset_sam(request):
    status = {"isresetsam": IS_RESET_SAM}
    return web.Response(
        text=json.dumps(status),
        content_type="application/json",
    )
